'''
    Author: Alexandre Manoury
    Python Version: 3.6
'''

import numpy as np

from exlab.interface.serializer import Serializable, Serializer

from dinos.data.data import SingleObservation
from dinos.data.space import Space


class UnboundedProperty(Exception):
    """
    Short summary.
    """
    pass


class Property(Serializable):
    """
    An abstract class corresponding to features and effectors
    """

    def __init__(self, entity, name, dim=None, absoluteName=None, visual=False):
        self.name = name
        self.absoluteName = absoluteName
        if self.absoluteName and entity.findAbsoluteName(self.absoluteName):
            raise Exception(f'An entity/property named {self.absoluteName} already exists within {entity.root.reference()}! \
                             Names should be uniques.')
        self.entity = entity
        self.entity.addProperty(self)

        self._dim = dim
        self.space = None
        self.discretization = 1.
        self.actions = {}

        self.learnable = True
        self.visual = visual

        if entity.activated:
            self._activate()

    def _activate(self):
        if self.space:
            return
        if not self._dim and self.observable():
            self._dim = len(self.observePlain())
        if self._dim is None:
            raise Exception(
                f'You should specify the dimension of the effector parameter space for {self}')
        self.space = Space(self.entity.manager, self._dim)
        self.space._property = self

    def _sid(self, serializer, raw=False):
        if self.absoluteName:
            return self.absoluteName
        return serializer.uid('property', f'{self.entity._sid(serializer, raw=True)}.{self.name}')

    def _serialize(self, serializer):
        dict_ = serializer.serialize(
            self, ['name', 'absoluteName', 'entity', 'dim'])
        return dict_

    @property
    def dim(self):
        return self._dim

    @property
    def discretizeStates(self):
        return self.entity.discretizeStates

    @property
    def discretizeActions(self):
        return self.entity.discretizeActions

    def observable(self):
        return False

    def controllable(self):
        return False

    @property
    def bounded(self):
        return self.space is not None

    # def bindSpace(self, spaces):
    #     fullname = self.fullname()
    #     if fullname[0] == '.':
    #         fullname = fullname[1:]
    #     spaces = [space for space in spaces if space.effector == fullname or space.observable == fullname]
    #     if spaces:
    #         self.space = spaces[0]

    # points
    def point(self, value, relative=None):
        return self.space.point(value, relative=relative)

    def goal(self, value, relative=None):
        return self.space.goal(value, relative=relative)

    def action(self, value, relative=None):
        return self.space.action(value, relative=relative)

    def zero(self, relative=None):
        return self.space.zero(relative=relative)

    def plainZero(self):
        return self.space.plainZero()

    def fullname(self):
        return self.entity.fullname() + "." + self.name

    def reference(self):
        if self.absoluteName:
            return f'#{self.absoluteName}'
        return f'{self.entity.reference(short=True)}.{self.name}'

    def icon(self):
        return '👁' if self.observable() else '' + '🕹' if self.controllable() else ''

    def __repr__(self):
        absName = f'#{self.absoluteName}' if self.absoluteName else ''
        return f"{self.icon()}{absName}'{self.entity.reference(short=True)}.{self.name}'"


class Effector(Property):
    """
    An actionable element
    """

    def __init__(self, entity, name, dim=1, absoluteName=None, actions={}):
        Property.__init__(self, entity, name, dim=dim,
                          absoluteName=absoluteName)
        self.learnable = False
        self.actions = {key: list(param) for key, param in actions.items()}

    def controllable(self):
        return True

    def performPlain(self, parameters):
        if not isinstance(parameters, list):
            parameters = [parameters]
        if len(parameters) != self._dim:
            raise Exception((f'Parameters size for {self} mismatch: expecting {self._dim} and received {len(parameters)} ' +
                             '(did you forget to specify the dim in the constructor?)'))
        self._performPlain(parameters)

    def _performPlain(self, parameters):
        pass

    def discrete(self):
        return self.entity.discreteActions

    def perform(self, action):
        if self.space is None:
            raise UnboundedProperty(
                f"{self.name} of {self.entity} is not bounded to any space")
        parameters = action.projection(self.space)
        self.performPlain(parameters.valueOrdered)


class Observable(Property):
    """
    A readable property
    """

    def __init__(self, entity, name, absoluteName=None, discretization=1., visual=False):
        Property.__init__(self, entity, name,
                          absoluteName=absoluteName, visual=visual)
        self.converter = None
        self.visualizer = None
        self.saver = "save"
        if discretization:
            self.discretization = discretization

    def observable(self):
        return True

    def observePlain(self):
        obs = self._observePlain()
        if not isinstance(obs, np.ndarray):
            if not isinstance(obs, list) and not isinstance(obs, tuple):
                obs = np.array([obs])
            else:
                obs = np.array(obs)
        if self.discretizeStates:
            obs = (obs // self.discretization).astype(np.int)
        return obs

    def discrete(self):
        return np.issubdtype(self.observePlain().dtype.type, np.integer)

    def _observePlain(self):
        return []

    def observe(self, formatParameters=None):
        if self.space is None:
            raise UnboundedProperty(
                f"{self.name} of {self.entity} is not bounded to any space")
        return SingleObservation(self.space, self.space.formatData(self.observePlain(),
                                                             formatParameters=formatParameters))

    @property
    def dim(self):
        assert len(self.observePlain()) == self._dim
        return self._dim

    def set_converter(self, converter_name):
        self.converter = converter_name

    def set_saver(self, saver_name):
        self.saver = saver_name

    def setVisualizer(self, visualizer):
        self.visualizer = visualizer

    def visualize(self, data=None, draw=True):
        if self.visualizer:
            self.visualizer(data=data, draw=draw)

    def convert_to_space(self, value):
        if self.converter:
            return getattr(self.entity.__class__, self.converter)(self.entity, value, to_feature=False)
        return value

    def convert_to_feature(self, value):
        if self.converter:
            return getattr(self.entity.__class__, self.converter)(self.entity, value, to_feature=True)
        return value

    def save(self):
        return getattr(self.entity.__class__, self.saver)(self.entity)


class MethodObservable(Observable):
    """
    A property provided by a given method
    """

    def __init__(self, entity, name, method, absoluteName=None, discretization=None, visual=False):
        self.method = method
        Observable.__init__(
            self, entity, name, absoluteName=absoluteName, discretization=discretization, visual=visual)

    def _observePlain(self):
        return self.method(property=self)


class FunctionObservable(Observable):
    """
    A property provided by a given function
    """

    def __init__(self, entity, name, function, absoluteName=None, discretization=None, visual=False, proxySpace=False):
        self.function = function

        if proxySpace:
            existingProperty = entity.propertyItem(name)

        Observable.__init__(
            self, entity, name, absoluteName=absoluteName, discretization=discretization, visual=visual)

        if proxySpace and existingProperty:
            self.space = existingProperty.space

    def _observePlain(self):
        return self.function(entity=self.entity, property=self)


class AttributeObservable(Observable):
    """
    A property directly read from an object attribute
    """

    def __init__(self, entity, name, attributeName, absoluteName=None, discretization=None, visual=False):
        self.attributeName = attributeName
        Observable.__init__(
            self, entity, name, absoluteName=absoluteName, discretization=discretization, visual=visual)

    def _observePlain(self):
        return getattr(self.entity, self.attributeName)


class MethodEffector(Effector):
    """
    An effector provided by a given method
    """

    def __init__(self, entity, name, method, dim, absoluteName=None, actions={}):
        self.method = method
        Effector.__init__(self, entity, name, dim=dim,
                          absoluteName=absoluteName, actions=actions)

    def _performPlain(self, parameters):
        return self.method(parameters, property=self)


class FunctionEffector(Effector):
    """
    An effector provided by a given function
    """

    def __init__(self, entity, name, function, dim, absoluteName=None, actions={}):
        self.function = function
        Effector.__init__(self, entity, name, dim=dim,
                          absoluteName=absoluteName, actions=actions)

    def _performPlain(self, parameters):
        return self.function(parameters, entity=self.entity, property=self)



'''
    File name: planner.py
    Author: Alexandre Manoury
    Python Version: 3.6
'''

from exlab.modular.module import Module
from exlab.utils.io import parameter

# from ..utils.debug import timethis
from dinos.utils.maths import linearValue
from dinos.data.data import Data, Goal, SingleAction, Action, ActionList
from dinos.data.path import ActionNotFound, Path, PathNode
from dinos.data.space import SpaceKind
# from dinos.models.model import Model
from dinos.utils.result import Result

from collections import namedtuple

import numpy as np
from scipy.spatial.distance import euclidean

import itertools
import random
import math
import copy


class PlannerModelMap(object):
    def __init__(self, model):
        self.model = model
        self.iteration = -1
        self.tree = None

    def get(self):
        if self.model.dataset.getIteration() != self.iteration:
            self.iteration = self.model.dataset.getIteration()


class PlanSettings(object):
    def __init__(self, computeActions=True, performing=False):
        self.computeActions = computeActions
        self.length = 0
        self.depth = -1
        self.context = False
        self.subPlanning = False
        self.computeSeperateState = False

        self.controlledSpaces = []
        self.dontMoveSpaces = []
        self.managedStateSpaces = set()

        self.allowContextPlanning = True
        self.forceContextPlanning = False
        self.tryContextPlanning = True
        self.mayUseContext = False
        self.hierarchical = None
        self.maxIterations = None

        self.performing = performing

        self.result = Result(None)

    def freeSpace(self, space):
        for s in self.controlledSpaces:
            if s.intersects(space):
                return False
        return True

    def clone(self, **kwargs):
        obj = copy.copy(self)
        obj.controlledSpaces = list(self.controlledSpaces)
        obj.dontMoveSpaces = list(self.dontMoveSpaces)
        obj.managedStateSpaces = set(self.managedStateSpaces)
        for key, value in kwargs.items():
            setattr(obj, key, value)
        return obj


Part = namedtuple('Part', ['model', 'goal'])


class Planner(Module):
    """
    Plans action sequences to reach given goals.
    This Planner can use both:
    - model hierarchy: to find primitive actions correponding to a goal
    - same model skill chaining: to reach out of range goals in a given model
    """

    MAX_DISTANCE = 0.01
    MAX_MOVE_UNCHANGED_SPACES = 1.
    GOAL_SAMPLE_RATE = 0.2

    def __init__(self, agent, hierarchical=None, chaining=None, options={}):
        super().__init__('planner', agent, loggerTag='plan')

        self.agent = agent
        self.environment = self.agent.environment
        self.dataset = self.agent.dataset

        self.options = options
        self.chaining = parameter(chaining, options.get('chaining', True))
        self.hierarchical = parameter(hierarchical, options.get('hierarchical', True))

        self.trees = {}
    
    @property
    def featureMap(self):
        return self.agent.featureMap
    
    def partitions(self, goal, settings=PlanSettings(), raiseException=True):
        settings = settings.clone()

        parts = goal.flat()
        space = goal.space

        self.logger.debug(f'Parting space to find models reaching {space}')

        for _ in range(1):
            # Find models related to our goal
            # models [(model, goal, related_parts, prob), ...]
            partsLeft = list(parts)
            partition = []
            while partsLeft:
                models = []
                for model in self.dataset.enabledModels():
                    spaces = list(model.reachesSpaces([p.space for p in partsLeft]))
                    if len(spaces) > 0 and settings.freeSpace(space):
                        models.append([model, spaces])

                if not models:
                    if raiseException:
                        raise ActionNotFound(
                            f'No model found planning how to reach {goal}')
                    return None
                if len(models) == 1:
                    bestModel = models[0]
                else:
                    models = np.array(models)
                    models = models[np.argsort([-model[0].competence() for model in models])]
                    bestModel = models[0]
                    # model = uniformRowSampling(models, [prob for _, _, prob in models])

                partsLeft = [p for p in partsLeft if p.space not in bestModel[1]]
                partition.append(
                    Part(bestModel[0], Goal(*[p for p in parts if p.space in bestModel[1]])))
        return partition
    
    def plannableSpace(self, space):
        partition = self.partitions(space.zero(), raiseException=False)
        return partition is not None

    def plan(self, goal, state=None, model=None, settings=PlanSettings()):
        return self.planDistance(goal, state=state, model=model, settings=settings)[:2]

    def planDistance(self, goal, state=None, model=None, settings=PlanSettings()):
        settings = settings.clone()
        settings.depth += 1

        goal = goal.convertTo(self.dataset, kind=SpaceKind.BASIC)
        if goal.space.primitive():
            raise Exception(f'Primitive actions such as {goal} cannot be goal, thus not planned!')

        if model:
            partition = [Part(model, goal)]
        else:
            partition = self.partitions(goal)
        # Using only 1 element from the partition
        partition = [partition[0]]

        totalDistance = 0
        paths = []
        for model, subgoal in partition:
            path, finalState, dist = self.__plan(
                model, subgoal, state=state, settings=settings)
            totalDistance += dist
            paths.append(path)

        return paths[0], finalState, totalDistance

    def planActions(self, actionList, state=None, model=None, settings=PlanSettings()):
        settings = settings.clone()
        settings.depth += 1

        # print('---- planActions')
        actionList = actionList.convertTo(self.dataset)
        nodes = []
        for action in actionList:
            node = PathNode(action=action)
            # print(action)
            # print(action.space.primitive())
            if not action.space.primitive():
                node.execution, _ = self.plan(
                    action, state=state, model=model, settings=settings)
            # print(node.paths)
            nodes.append(node)
        # print('----')
        return Path(nodes)

    def __plan(self, model, goal, state=None, settings=PlanSettings()):
        planning = Planning(self, model, goal, state, settings)
        return planning.execute()[:3]

    # def _contextPlanning(self, model, goal, state=None, settings=PlanSettings()):
    #     settings = settings.clone()
    #     path = []
    #     print(model.controllableContext())
    #     if not model.controllableContext():
    #         return None, None, None

    #     c0 = model.adaptContext(goal, state.context())
    #     if not c0:
    #         return None, None, None

    #     try:
    #         cpath, _, cdistance = self.planDistance(
    #             c0, state=state, settings=settings)
    #     except Exception:
    #         return None, None, None
    #     print('--------')
    #     print(cpath[0][-1].state)
    #     print(c0)

    #     return c0, cpath, cdistance


class Planning(object):
    DEBUG = {}

    def __init__(self, planner, model, goal, state, settings):
        self.DEBUG['plannings'] = self.DEBUG.get('plannings', 0) + 1
        self.planner = planner
        self.settings = settings

        self.logger.debug(
            f'=== New planning {self.logTag()} === -> {goal} using {model}')# with context {state}

        self.model = model
        self.space = model.outcomeSpace
        self.space._validate()

        self.startState = parameter(state, self.planner.environment.state(self.planner.dataset, self.planner.featureMap))
        self.startPosition = self.startState.context().projection(self.space)
        self.goal = goal.projection(self.space).relativeData(self.startState)
        self.checkSpaceIsControllable(self.goal, self.space)

        self.logger.debug(
            f'== Relative goal is {self.goal} (starting pos {self.startPosition}) {self.startState}')

        self.initParameters()
        self.initVariables()

    def logTag(self):
        return f'(d{self.settings.depth})'

    @property
    def logger(self):
        return self.planner.logger

    def initParameters(self):
        ratio = self.space.number / 200
        self.maxIter = parameter(self.settings.maxIterations, int(linearValue(10, 150, ratio)))
        self.maxIterNoNodes = int(linearValue(10, 20, ratio))

        self.maxDistanceBest = self.model.getPrecision(Planner.MAX_DISTANCE, 0.5)
        self.maxDistance = self.maxDistanceBest * 3 + 0.02 * self.goal.norm()
        self.maxDistanceIncomplete = self.maxDistanceBest * 8 + 0.1 * self.goal.norm()

        if self.settings.dontMoveSpaces:
            self.maxIter *= 2
            self.maxIterNoNodes *= 2
        if self.settings.subPlanning:
            self.maxIter = min(self.maxIter, 2)
            self.maxIterNoNodes = min(self.maxIterNoNodes, 2)
            self.maxDistanceIncomplete = self.maxDistance
        
        self.excludeStateSpaces = set(self.settings.managedStateSpaces)
        # self.managedStateSpaces = self.planner.dataset.controlledSpaces([self.model]) - self.settings.managedStateSpaces
        self.settings.managedStateSpaces |= self.planner.dataset.controlledSpaces([self.model])

        self.contextPlanningPossible = True
        if not self.model.contextSpace:
            self.contextPlanningPossible = False
        else:
            contextSpace = self.model.contextSpace.convertTo(kind=SpaceKind.BASIC)
            controlled = not self.settings.freeSpace(contextSpace)
            if controlled or not self.planner.plannableSpace(contextSpace):
                self.contextPlanningPossible = False
                self.logger.debug(
                    f'{self.logTag()} Context planning wont be available as {self.model.contextSpace} is {"already controlled" if controlled else "uncontrollable"}')
    
    def initVariables(self):
        self.minDistanceReached = self.goal.norm()
        self.closestNodeToGoal = None
        self.attemptCloseToGoal = 0

        self.nodes = [PathNode(pos=self.space.zero(), absPos=self.startPosition, state=self.startState)]

        self.path = Path()
        self.directGoal = True
        self.currentContextFailures = 0
        self.lastPosition = None

        # 
        self.invalidPointRatioLimit = 0.05
        self.ignoreConstraintDistanceLimit = 0.05 * self.space.maxDistance

        self.attemptedUnreachable = []
        self.attemptedBreakConstraint = []
        self.lastInvalids = []

        # Attempt results
        self.a0, self.y0, self.c0, self.distance0 = [None, None, None, None]

    def execute(self):
        if self.settings.subPlanning:
            return self.executeSingleStep()
        else:
            return self.executeMultiStep()

    def executeMultiStep(self):
        if self.space._number < 20 or self.minDistanceReached < self.maxDistanceBest:
            return Path(), None, 0, None

        # Main loop
        self.i = -1
        self.iterationOffset = 0
        while self.i < self.maxIter + self.iterationOffset:
            self.DEBUG['iterations'] = self.DEBUG.get('iterations', 0) + 1
            self.i += 1

            if len(self.nodes) <= 1 and self.i > self.maxIterNoNodes or len(self.nodes) <= 2 and self.i > self.maxIterNoNodes * 2:
                self.logger.warning(f'Not enough nodes, aborting...')
                break
    
            # Generating distant subgoals
            self.subGoal, useGoal = self.generateSubGoal()
        
            nearestNode = self.nearestNode(self.subGoal, useGoal)
            if nearestNode is None:
                self.iterationOffset += 1
                continue
            self.state = nearestNode.state
            context = self.state.context()

            self.logger.debug2(
                f'{self.logTag()} Iter {self.i}: chosen subGoal is {self.subGoal} (direct={self.directGoal}, final goal {self.goal})\n    Nearest node is: {nearestNode.pos}')#\n    With context {self.state}

            baseMove = self.subGoal - nearestNode.pos

            reachable, move = self.searchWithCurrentContext(baseMove, context)
            if reachable is None:
                self.logger.debug2(
                    f'{self.logTag()} Iter {self.i}: invalid move {self.y0}, skipping iteration')
                self.directGoal = False
                self.iterationOffset += 1
                if move is True:
                    self.currentContextFailures += 1
                if useGoal:
                    nearestNode.cantConnectToGoal = True
                continue
            reachable, contextPath = self.searchWithDifferentContext(baseMove, reachable)

            childPath = None
            if reachable:
                reachable, childPath = self.checkControllableHierarchy(self.a0)
                self.logger.debug(f'{self.logTag()} Iter {self.i}: hierarchical check: move {"possible" if reachable else "IMPOSSIBLE"}')

            # self.logger.info(f'{self.logTag()} Iter {self.i}: end {reachable}')
            if not reachable:
                self.directGoal = False
                self.logger.debug2(
                    f'{self.logTag()} Iter {self.i}: not reachable!!')
                nearestNode.failures += 1
                if move:
                    self.attemptedUnreachable.append((nearestNode.pos + move).plain() + nearestNode.pos.plain())
                    self.lastInvalids.append((nearestNode.pos + move).plain())
                if useGoal:
                    nearestNode.cantConnectToGoal = True
                continue
        
            newNode = self.addNode(nearestNode, contextPath, useGoal, childPath)
            if not newNode:
                continue

            if self.checkDistance(newNode):
                break
        
        if self.closestNodeToGoal and not self.path:
            self.path = self.closestNodeToGoal.createPath(self.goal, self.pathSettings())

        if self.logger.isDebugging() and not self.settings.subPlanning:
            self._plotNodes(self.nodes, self.goal, self.space, self.attemptedUnreachable, self.attemptedBreakConstraint)

        self.logger.debug(
            f"{self.logTag()} Planning {'generated' if self.path else 'failed'} for goal {self.goal} in {self.i+1} step(s)")
        
        # if self.path:
            # self.path = self.simplifyPath(self.path)
            # self._plotNodes(self.path.nodes, self.goal, self.space, self.attemptedUnreachable, self.attemptedBreakConstraint)
        
        if not self.settings.performing and not self.settings.context and self.settings.depth == 0:
            self.settings.result.planningSuccess = self.path is not None
            self.settings.result.planningSteps = self.i

        minDistance = max(0, self.minDistanceReached)

        # if not self.path and minDistance > 0:
        #     self.model.updatePrecision(False, min(minDistance, self.space.maxDistance * 0.2))

        finalState = self.path[-1].state if self.path else None
        return self.path, finalState, minDistance, None
    
    def executeSingleStep(self):
        if self.space._number < 20:
            return Path(), None, 0, None

        self.i = 0
        self.state = self.startState
        node = None
        reachable = self.attempt(self.goal, self.state.context())
        # self.logger.warning(f'{self.goal} {self.y0}')

        if reachable:
            reachable, childPath = self.checkControllableHierarchy(self.a0)

        if reachable:
            node = self.addNode(self.nodes[0], None, True, childPath)

        if not node:
            return Path(), None, 0, None

        path = node.createPath(self.goal, self.pathSettings())
        finalState = path[-1].state if path else None
        # print('===')
        # print(path)
        # print(finalState)
        return path, finalState, euclidean(node.pos.plain(), self.goal.plain()), None

    def searchWithCurrentContext(self, baseMove, context):
        riskyMove, move, safeMove, _, _ = self.findBestMove(baseMove, context)

        maxIncorrectMoveDistance = 0.01 * self.space.maxDistance * self.model.limitMoves
        maxNoMoveDistance = 0.15 * self.space.maxDistance * self.model.limitMoves
        if self.minDistanceReached < self.space.maxDistance:
            maxNoMoveDistance *= self.minDistanceReached / self.space.maxDistance
        maxNoMoveDistanceMin = maxNoMoveDistance * 2

        # self.logger.error(
        #     f'{self.logTag()} Iter {self.i}: {maxNoMoveDistance} {maxNoMoveDistanceMin} {self.minDistanceReached}')

        if move is None:
            self.logger.error(
                f'{self.logTag()} Iter {self.i}: failed to find a move!\n{self.model} {self.goal} {context}')
            return None, None

        if move.norm() < self.space.maxDistance * 0.001:
            return False, None

        self.logger.debug2(
            f'{self.logTag()} Iter {self.i}: chosen move is {move} (risky move is {riskyMove}) {maxNoMoveDistance} {maxNoMoveDistanceMin}')

        reachable = False
        if not self.settings.forceContextPlanning:
            if riskyMove:
                if riskyMove.norm() < maxNoMoveDistance and self.minDistanceReached > maxNoMoveDistanceMin:
                    return None, True
                reachable = self.attempt(riskyMove, context) if riskyMove else False

            if not reachable:
                if move.norm() < maxNoMoveDistance and self.minDistanceReached > maxNoMoveDistanceMin:
                    return None, True
                reachable = self.attempt(move, context)
            # if reachable:
            #     newPosition = nearestNode.pos + y0
            #     absPos = startPos + newPosition
            #     dist = newPosition.distanceTo(attemptMove)

            if not reachable:
                self.currentContextFailures += 1

            if not reachable and safeMove:
                reachable = self.attempt(safeMove, context)
            
            for multiplier in (1., 0.5):
                if not reachable:
                    reachable = self.attempt(move * multiplier, context)

            if reachable and self.y0.norm() < maxIncorrectMoveDistance and self.minDistanceReached > maxIncorrectMoveDistance * 2:
                return False, None
            if reachable and self.y0.norm() < maxNoMoveDistance and self.minDistanceReached > maxNoMoveDistanceMin:
                return None, None
            
            # if reachable and baseMove.norm() > 0.01:
            #     if self.y0.norm() < 0.01:
            #         reachable = False
            #     else:
            #         # diff = (subGoal - y0).norm() / subGoal.norm()
            #         orientation = 1. - baseMove.npPlain().dot(self.y0.npPlain()) / (baseMove.norm() * self.y0.norm())
            #         if orientation > 0.4: # or y0.norm() < 0.2 * goal.norm() or y0.norm() > 5.0 * goal.norm():
            #             reachable = False
            #             self.logger.info(
            #                 f'{self.logTag()} Iter {self.i}: Move orientation is off compared to baseMove!\nOrt: {orientation}, y0: {self.y0}, baseMove: {baseMove}, move: {move}')
        # self.logger.info(f'{self.logTag()} Iter {self.i}: {reachable}')
        return reachable, move
    
    def searchWithDifferentContext(self, baseMove, reachableWithContext):
        reachable = reachableWithContext
        contextPath = None
        settingsTestContext = self.settings.forceContextPlanning or self.settings.tryContextPlanning
        if (settingsTestContext or not reachableWithContext) and self.model.contextSpace and self.contextPlanningPossible:
            # Save with context results
            a0wc, y0wc, d0wc = [self.a0, self.y0, self.distance0]

            riskyMove, move, safeMove, nearestMove, c0 = self.findBestMove(baseMove)
            c0 = self.findBestContext(move)
            riskyMove, move, safeMove, nearestMove, _ = self.findBestMove(baseMove, c0)

            # c0 = self.model.contextSpace.getPoint(nearestMoveId)[0]
            if settingsTestContext:
                self.logger.debug2(f'{self.logTag()} Iter {self.i}: testing with context {c0}')
            else:
                self.logger.debug2(f'{self.logTag()} Iter {self.i}: move still not reachable, trying to use a different context {c0}')

            reachable = self.attempt(riskyMove, c0, adaptContext=True) if riskyMove else False

            if not reachable:
                reachable = self.attempt(move, c0, adaptContext=True)
            # if reachable:
            #     newPosition = nearestNode.pos + y0
            #     absPos = startPos + newPosition
            #     dist = newPosition.distanceTo(attemptMove)

            if not reachableWithContext or self.currentContextFailures >= self.maxIter * 0.1:
                if not reachable and safeMove:
                    reachable = self.attempt(safeMove, c0, adaptContext=True)
                
                for multiplier in (1., 0.5, 0.4, 0.25, 0.15):
                    if not reachable:
                        reachable = self.attempt(move * multiplier, c0, adaptContext=True)

            # print('===')
            # print(reachable, y0, a0)

            switchBackWithContext = False
            if reachableWithContext and (not reachable or d0wc <= self.distance0 + self.maxDistanceBest):
                switchBackWithContext = True
            elif reachable:
                self.logger.debug(
                    f'{self.logTag()} Iter {self.i}: choose to {"change" if reachableWithContext else "use current"} context ({d0wc} > {self.distance0 + self.maxDistanceBest}), is the context reachable?')

                newSettings = self.settings.clone(context=True)
                # if self.i == 0:
                newSettings.dontMoveSpaces.append(self.space)
                # First context planning
                # if i == 0:
                #     newSettings.maxIterations = 100
                contextPath, contextState = self.attemptPlanningContext(c0, self.state, newSettings)

                # Try without dontMoveSpaces constraint
                # if not contextPath:
                #     newSettings.dontMoveSpaces = []
                #     contextPath, contextState = self.attemptPlanningContext(c0, self.state, newSettings)

                if contextPath:
                    c0 = contextState.context()
                    reachable = self.attempt(nearestMove, c0)

                    if not reachable:
                        reachable = self.attempt(nearestMove * 0.5, c0)
                else:
                    reachable = False

                if not reachable:
                    switchBackWithContext = True
                    self.logger.debug(
                        f'{self.logTag()} Iter {self.i}: context {c0} is not reachable')
                elif reachableWithContext and d0wc <= self.distance0 + self.maxDistanceBest * 0.5 * len(contextPath):
                    switchBackWithContext = True
            
            if switchBackWithContext and reachableWithContext:
                self.logger.debug2(
                    f'{self.logTag()} Iter {self.i}: switching back to not modifying the context')
                reachable, self.a0, self.y0, self.distance0 = [True, a0wc, y0wc, d0wc]
                contextPath = None

            if contextPath:
                self.state = contextState
                self.logger.debug(
                    f'{self.logTag()} Iter {self.i}: changing context to reach c0={c0} with a {len(contextPath)} step sub path')

        return reachable, contextPath
    
    def addNode(self, nearestNode, contextPath, useGoal=False, childPath=None):
        newPosition = nearestNode.pos + self.y0
        absPosition = self.startPosition + newPosition
        goalDistance = euclidean(newPosition.plain(), self.goal.plain())

        # Creating a new node
        newState = self.nextState(nearestNode, goalDistance, self.state, useGoal)
        if newState is None:
            return None

        self.logger.debug2(
            f'{self.logTag()} Iter {self.i}: move is usable ({self.a0}->{self.y0}), future state: {newState} ({absPosition})')

        self.lastInvalids = []
        # if len(lastInvalids) > 5:
        #     lastInvalids = lastInvalids[-5:]
        # if lastInvalids:
        #     lastInvalids = lastInvalids[1:]

        newNode = PathNode(pos=newPosition, absPos=absPosition, action=self.a0, goal=self.y0, model=self.model, parent=nearestNode,
                            state=newState)
        newNode.execution = childPath
        newNode.context = contextPath
        newNode.ty0 = self.model.npForward(self.a0, self.state.context())
        # nodePositions[len(nodes)] = newPosition.npPlain()
        self.nodes.append(newNode)
        self.logger.debug2(
            f'{self.logTag()} Iter {self.i}: node {newNode} attached to {nearestNode}')
        return newNode

    def nextState(self, nearestNode, goalDistance, state, useGoal=False):
        self.DEBUG['next'] = self.DEBUG.get('next', 0) + 1

        newState = state.copy().apply(self.a0, self.excludeStateSpaces)

        if self.settings.dontMoveSpaces and goalDistance > self.ignoreConstraintDistanceLimit:
            if self.hasSpacesChanged(nearestNode, newState):
                self.directGoal = False
                nearestNode.failures += 1
                self.attemptedBreakConstraint.append((nearestNode.pos + self.y0).plain() + nearestNode.pos.plain())
                self.lastInvalids.append((nearestNode.pos + self.y0).plain())
                if useGoal:
                    nearestNode.cantConnectToGoal = True
                return None
        return newState

    def checkDistance(self, newNode):
        # Distance
        distance = euclidean(newNode.pos.plain(), self.goal.plain())
        self.logger.debug2(
            f'{self.logTag()} Iter {self.i}: distance from goal {distance:.3f} (max {self.maxDistance:.3f})')

        # print(distance, self.minDistanceReached,
        #       self.maxDistanceBest, self.maxDistance, newNode.pos.plain(), self.goal.plain())
        if distance > self.minDistanceReached:
            self.directGoal = False
            if self.minDistanceReached < self.maxDistance and self.closestNodeToGoal:
                self.attemptCloseToGoal += 1
                if self.attemptCloseToGoal >= 3:
                    self.logger.debug(f'{self.logTag()} Iter {self.i}: close enough to the goal {self.minDistanceReached:.3f} (max {self.maxDistance:.3f})')
                    self.path = self.closestNodeToGoal.createPath(self.goal, self.pathSettings())
                    if not self.settings.performing and not self.settings.context and self.settings.depth == 0:
                        self.settings.result.planningDistance = (self.minDistanceReached, self.maxDistance)
                    return True
            else:
                self.attemptCloseToGoal = 0
        else:
            self.minDistanceReached = distance
            self.attemptCloseToGoal = 0

        # print(distance)
        if distance < self.maxDistanceBest:
            self.logger.debug(f'{self.logTag()} Iter {self.i}: close enough to the goal {distance:.3f} (max best {self.maxDistanceBest:.3f})')
            self.path = newNode.createPath(self.goal, self.pathSettings())
            if not self.settings.performing and not self.settings.context and self.settings.depth == 0:
                self.settings.result.planningDistance = (distance, self.maxDistanceBest)
            return True
        elif distance < self.maxDistanceIncomplete:
            self.closestNodeToGoal = newNode
    
        # newPosition = nearestNode.pos + y0
        # if lastPos and np.sum(np.abs(newPosition.npPlain() - lastPos)) < 1.:
        #     self.logger.debug2(
        #         f'{self.logTag()} Iter {self.i}: deadlock, we are not moving enough! Stopping here')
        #     break
        self.lastPosition = newNode.pos.plain()
        return False
    
    def simplifyPath(self, path):
        self.i = 0
        nodes = path.nodes
        index = 0
        maxJump = 4
        while index < len(nodes) - 2:
            node = path[index]
            state = node.parent.state
            self.logger.debug2(f'Simplify: trying a jump from {index}/{len(path)}')
            for jump in range(maxJump, 0, -1):
                if index + jump >= len(path) - 1:
                    continue
                # check if there is a context change
                if any(path[index + i].context for i in range(jump + 2)):
                    self.logger.debug2(f'Simplify: context!')
                    continue

                self.logger.debug2(f'Simplify: trying a jump +{jump} from {index}/{len(path)}')

                # Pre move
                preMove = node.goal
                for i in range(1, jump + 1):
                    preMove += path[index + i].goal
                if not self.attempt(preMove, state.context()):
                    continue
            
                newPosition = node.parent.pos + self.y0
                absPosition = self.startPosition + newPosition
                goalDistance = euclidean(newPosition.plain(), self.goal.plain())
                newState = self.nextState(node.parent, goalDistance, state)
                if newState is None:
                    continue
                newPreNode = PathNode(pos=newPosition, absPos=absPosition, action=self.a0, goal=self.y0, model=self.model, parent=node.parent,
                                      state=newState)
                
                self.logger.debug2(f'Simplify: pre jump possible, reaching {newPosition} instead of {path[index + jump].pos} by doing {self.a0}=>{self.y0}')

                # Post Move
                # preA0 = self.a0
                # preY0 = self.y0
                postMove = path[index + jump + 1].goal + preMove - self.y0
                if not self.attempt(postMove, state.context()):
                    continue

                newPosition = newPreNode.pos + self.y0
                absPosition = self.startPosition + newPosition
                goalDistance = euclidean(newPosition.plain(), self.goal.plain())
                newState = self.nextState(newPreNode, goalDistance, newPreNode.state)
                if newState is None:
                    continue
                newPostNode = PathNode(pos=newPosition, absPos=absPosition, action=self.a0, goal=self.y0, model=self.model, parent=newPreNode,
                                       state=newState)
                
                self.logger.debug2(f'Simplify: post jump possible, reaching {newPosition} instead of {path[index + jump + 1].pos} by doing {self.a0}=>{self.y0} ({path[index + jump + 1].goal})')
                self.logger.debug2(f'Simplify: compacting {jump+1}->2')

                nodes = nodes[0:index] + [newPreNode, newPostNode] + nodes[index + jump + 1:]
                break
            index += 1

        return path

    def attempt(self, move, context, adaptContext=False):
        self.DEBUG['attempts'] = self.DEBUG.get('attempts', 0) + 1
        precision = 0.45 + (self.i / self.maxIter) * 0.05
        precisionOrientation = 0.05 + (self.i / self.maxIter) * 0.05
        if self.settings.subPlanning:
            precision = 0.55
            precisionOrientation = 0.1

        if adaptContext:
            context = self.findBestContext(move)

        for again in range(2):
            reachable, self.a0, self.y0, self.distance0 = self.model.reachable(
                move, context=context, precision=precision, precisionOrientation=precisionOrientation, dontMove=self.settings.dontMoveSpaces)
            
            if not reachable and not again and self.model.contextSpace:
                pass
                # settings = self.settings.clone(subPlanning=True)
                # path, state = self.planner.plan(controllable, self.state, settings=settings)

        contextLog = f' with context {context}' if adaptContext else ''
        self.logger.debug2(f'{self.logTag()} Iter {self.i}: testing {move}{contextLog}: {"reachable" if reachable else "not reachable"}')
        # print(f'{self.logTag()} Iter {self.i}: testing {move}->{self.y0}{contextLog}: {"reachable" if reachable else "not reachable"}')
        return reachable
    
    def attemptPlanningContext(self, context, state, settings):
        try:
            return self.planner.plan(context.setRelative(False), state, settings=settings)
        except ActionNotFound:
            return None, None

    def checkControllableHierarchy(self, controllable):
        if controllable.space.primitive():
            return True, None
        try:
            settings = self.settings.clone(subPlanning=True)
            path, state = self.planner.plan(controllable, self.state, settings=settings)
            if path:
                self.state = state
                # self.logger.warning(f'{self.a0}')
                self.a0 = path[-1].pos
                # self.logger.warning(f'{self.a0}')
            return bool(path), path
        except ActionNotFound:
            return False, None
    
    def checkSpaceIsControllable(self, goal, space):
        if not self.settings.freeSpace(space):
            error = f'Failed to create a path to reach {goal}. Space {space} trying to be controlled twice! (controlled spaces: {self.settings.controlledSpaces})'
            self.logger.debug(error)
            raise ActionNotFound(error, None)
        self.settings.controlledSpaces += space
    
    def pathSettings(self):
        settings = self.settings.clone()
        settings.controlledSpaces.remove(self.space)
        return settings
    
    def hasSpacesChanged(self, nearestNode, newState):
        difference = newState.difference(nearestNode.state)
        for space in self.settings.dontMoveSpaces:
            spaceDifference = difference.projection(space)
            if spaceDifference.norm() > self.planner.MAX_MOVE_UNCHANGED_SPACES:
                # self.logger.warning(f'{newState.context().projection(space)} vs {state.context().projection(dmSpace)} by doing {a0}')
                self.logger.debug2(
                    f'{self.logTag()} Iter {self.i}: move is reachable but it affects {space}: {spaceDifference} that shouldnt be changed')
                return True
        return False
    
    def generateSubGoal(self):
        # Finding a subgoal
        useGoal = False
        if random.uniform(0, 1) <= self.planner.GOAL_SAMPLE_RATE or self.directGoal:
            subGoal = self.goal
            useGoal = True
        else:
            subGoal = self.generateRandomPoint()
            # space.goal([random.uniform(minrand, maxrand) for x in range(space.dim)])

            if self.lastInvalids:
                invalids = np.array(self.lastInvalids)
                for _ in range(100):
                    distanceToNearestNode = (self.nearestNode(self.subGoal).pos - subGoal).norm()
                    dists = np.sum((invalids - subGoal.npPlain()) ** 2, axis=1) ** 0.5
                    if not np.any(dists < self.invalidPointRatioLimit * self.space.maxDistance) and distanceToNearestNode > 0.2 * self.space.maxDistance:
                        break
                    subGoal = self.generateRandomPoint()
                    useGoal = False
                    # self.logger.debug2(f'(d{settings.depth}) Iter {i}: Too close to invalid point, retrying...')
        return subGoal, useGoal
    
    def generateRandomPoint(self):
        origin = random.uniform(0., 1.) * self.goal.npPlain()
        # point = [random.uniform(0., 1.) for _ in range(space.dim)]
        # point = point * self.goal
        width = 2. * min(0.05 * self.space.maxDistance + self.goal.norm(), self.space.maxDistance)
        point = origin + width * np.array([random.uniform(-1., 1.)
                                   for _ in range(self.space.dim)])
        return self.space.goal(point)

    def nearestNode(self, point, useGoal=False):
        pointPlain = point.plain()
        validNodes = [node for node in self.nodes if not node.cantConnectToGoal or not useGoal]
        if not validNodes:
            return None
        dlist = np.sum((np.array([node.pos.plain() for node in validNodes]) - pointPlain) ** 2, axis=1) ** 0.5
        dlist = dlist * np.array([node.penalty() for node in validNodes])
        return validNodes[np.argmin(dlist)]

    def findBestMove(self, baseMove, context=None):
        # lastVariance = -1.
        # We are looking for a valid point in the outcome space
        # self.logger.debug2(f'=====================')

        ratio = (self.space._number - 100) / 300
        number = int(linearValue(1, 2, ratio))
        close = linearValue(1., 0.05, ratio)
        riskyMove, _, _ = self._searchBestMove(baseMove, number, close, context)
        if riskyMove is None:
            return None, None, None, None, None
        
        number = int(linearValue(1, 6, ratio))
        close = linearValue(0.2, 0.05, ratio)
        move, nearestMove, ids = self._searchBestMove(baseMove, number, close, context)
        if move is None:
            return None, None, None, None, None
        
        if riskyMove.norm() < move.norm():
            riskyMove = None

        c0 = None
        if context is None:
            cs = self.model.contextSpace.getNpPlainPoint(ids)
            c0 = self.model.contextSpace.asTemplate(np.mean(cs, axis=0))

        safeMove = move
        if (move - baseMove).norm() < move.norm() * 0.1:
            move = baseMove
        else:
            move *= 0.9
        # self.logger.debug2(f'---------------------')
        return riskyMove, move, safeMove, nearestMove, c0  # , nearestMoveId
    
    def findBestContext(self, move):
        # ratio = (self.space._number - 100) / 300
        number = 1 #int(linearValue(1, 6, ratio))

        ids, _, _ = self.model.nearestOutcome(move, n=number, nearestUseContext=False)
        cs = self.model.contextSpace.getNpPlainPoint(ids)
        c0 = self.model.contextSpace.asTemplate(np.mean(cs, axis=0))
        # self.logger.debug2(
        #     f'HEY {ids} {self.model.outcomeSpace.getNpPlainPoint(ids)}')
        return c0
    
    def _searchBestMove(self, baseMove, number, close, context):
        move = baseMove.clone()
        for j in range(10):
            if j > 0:
                move *= 1. * distanceToCenter / move.norm()
            if j == 1:
                move *= self.model.limitMoves
            # self.logger.debug2(
            #     f'(d{settings.depth}) Iter {i} {j}: {move}')

            ids, _, _ = self.model.nearestOutcome(
                move, context=context, n=number, nearestUseContext=False)
            # ids, dists = self.model.outcomeContextSpace.nearestDistance(moveContext, n=5, restrictionIds=ids)
            if len(ids) == 0:
                return None
            nearestMoveId = ids[0]
            nearestMove = self.space.getPoint(nearestMoveId)[0]

            points = self.space.getNpPlainPoint(ids)
            center = np.mean(points, axis=0)
            distanceToCenter = np.sum(center ** 2) ** .5
            distanceMoveToCenter = np.mean(
                np.sum((center - move.npPlain()) ** 2) ** .5)
            # self.logger.debug2(f'(d{self.settings.depth}) Iter {i} {j}: {self.space.getPlainPoint(ids)}')
            # self.logger.debug2(f'(d{self.settings.depth}) Iter {i} {j}: {nearestMove} {distanceToCenter} {distanceToNode}')

            # if context is None:
            #     cs = self.model.contextSpace.getNpPlainPoint(ids)
            #     c0 = np.mean(cs, axis=0)
            # self.logger.info(f'{j}: {move} {distanceToCenter} {distanceMoveToCenter} {self.space.maxDistance * close}\n    {points}')

            if (j > 0 and distanceMoveToCenter <= self.space.maxDistance * close) or distanceToCenter > move.norm():
                break
        
        return move, nearestMove, ids

    def _plotNodes(self, nodes, goal, space, attemptedUnreachable, attemptedBreakConstraint):
        import matplotlib.pyplot as plt
        goalPlain = goal.plain()
        # plt.figure()
        points = np.array([self.generateRandomPoint().plain() for _ in range(1000)])
        plt.scatter(points[:, 0], -points[:, 1], marker='.', color='gray')
        if attemptedUnreachable:
            # attemptedUnreachable = np.array(attemptedUnreachable)
            # plt.scatter(attemptedUnreachable[:, 0], -attemptedUnreachable[:, 1], marker='.', color='black')
            # for x, y, fx, fy in attemptedUnreachable:
            for x, y, fx, fy in attemptedUnreachable:
                plt.plot([fx, x], [-fy, -y], 'm,--')
        if attemptedBreakConstraint:
            # attemptedBreakConstraint = np.array(attemptedBreakConstraint)
            # plt.scatter(attemptedBreakConstraint[:, 0], -attemptedBreakConstraint[:, 1], marker='o', color='purple')
            for x, y, fx, fy in attemptedBreakConstraint:
                plt.plot([fx, x], [-fy, -y], 'r,--')
        plt.scatter(goalPlain[0], -goalPlain[1], marker='X', color='orange')
        for node in nodes:
            if node.parent is not None:
                pos = node.pos.plain()
                parentPos = node.parent.pos.plain()
                plt.scatter(pos[0], -pos[1], marker='+', color='green' if node.valid else 'blue')
                plt.plot([parentPos[0], pos[0]], [-parentPos[1], -pos[1]], f"{'g' if node.valid else 'b'},-")
        for node in nodes:
            if node.parent is None:
                pos = node.pos.plain()
                plt.scatter(pos[0], -pos[1], marker='x', color='blue')
        for node in nodes:
            if node.parent is not None:
                pos = node.pos.plain()
                if node.cantConnectToGoal:
                    plt.scatter(pos[0], -pos[1], marker='X', color='red')
        #plt.scatter(goalPlain[0], goalPlain[1], 'x', color='purple')
        #plt.draw()
        plt.show()




from exlab.interface.serializer import Serializable

from dinos.representation.entity import Entity, ProxyEntity


class EntityManager(object):
    def __init__(self, name, entityCls=Entity, proxyCls=ProxyEntity):
        self.name = name
        self.entityCls = entityCls
        self.proxyCls = proxyCls

        self.engine = None
        self.world = self.entityCls('root', manager=self)

        self.conserveRoot = False

    def __repr__(self):
        return f'EntityManager({len(self.world.cascadingChildren()) + 1} entities)'
    
    def findEntity(self, name):
        return self.world.cascadingChild(name)

    def findProperty(self, name):
        return self.world.cascadingProperty(name)
    
    def entities(self):
        return [self.world] + self.world.cascadingChildren()
    
    def agents(self):
        return [host.hosting for host in self.world.hosts() if host.hosting]

    def scheduledActions(self, notNone=True):
        if notNone:
            return {agent: agent.scheduled for agent in self.agents() if agent.scheduled}
        return {agent: agent.scheduled for agent in self.agents()}

    def countScheduledActions(self):
        return len(self.scheduledActions())

    def clear(self):
        self.world.clearChildren()
    
    def propertySpace(self, filterOrProperty=None, kind=None, dataset=None):
        if isinstance(filterOrProperty, str):
            filterOrProperty = self.world.cascadingProperty(filterOrProperty)
        space = filterOrProperty.space
        if dataset:
            space = space.convertTo(spaceManager=dataset, kind=kind)
        return space

    # Entities
    def convertEntity(self, entity):
        relatedEntity = next((e for e in self.entities() if e.linkedTo(entity)), None)
        if relatedEntity:
            return relatedEntity

        entity = entity.createLinkedEntity(self)
        if entity.isRoot() and not self.conserveRoot:
            self.world = entity
        elif not entity.parent:
            self.world.addChild(entity)
        entity = self.customEntityConversion(entity)
        return entity

    def customEntityConversion(self, entity):
        return entity



import copy
import random
import numpy as np

from exlab.modular.module import Module

from exlab.utils.io import parameter
from dinos.utils.move import MoveConfig
from dinos.utils.maths import linearValue

# from ...data.dataset import Action
# from dinos.data.data import *
from dinos.data.space import SpaceKind
from dinos.data.event import InteractionEvent
from dinos.data.path import ActionNotFound



class Strategy(Module):
    """Strategy usable by a learning agent."""
    DISTANCE_MAX_CONTEXT = 0.05

    def __init__(self, agent, name=None, performer=None, planner=None, options={}):
        """
        name string: name of the strategy
        """
        self.name = name if name else (self.__class__.__name__[
                                       :1].lower() + self.__class__.__name__[1:])
        super().__init__(f'Strategy {self.name}', agent, loggerTag='strategy')

        self.agent = agent
        self.options = options
        # self.iterations = options.get('iterations', 1)

        # Short term memory containing all actions & outcomes reached during the last learning episode
        self.memory = []
        # self.config = None
        # self.n = 0

        # self.complex_actions = complex_actions
        # self.resetEnv = True

        self.performer = parameter(performer, self.agent.performer)
        self.planner = parameter(planner, self.agent.planner)
        # self.addChildModule(self.performer)

    def __repr__(self):
        return f'Strategy {self.name}'
    
    def _sid(self, serializer):
        return serializer.uid('strategy', self.name)

    def _serialize(self, serializer):
        dict_ = super()._serialize(serializer)
        dict_.update(serializer.serialize(self, ['name']))
        return dict_

    # @classmethod
    # def _deserialize(cls, dict_, agent, options=None, obj=None):
    #     obj = obj if obj else cls(agent, dict_.get(
    #         'name'), options=dict_.get('options', {}))
    #     obj = Module._deserialize(dict_, options=options, obj=obj)
    #     return obj

    def available(self, space):
        """Says if the strategy is available to the agent."""
        return True

    def trainable(self):
        return True

    def testable(self):
        return True

    def run(self, config):
        """Runs the strategy in train or test mode (function used when the strategy does not require a goal/task)."""
        self._preRun(config)
        while len(self.memory) < config.iterations:
            self._run(config)
        return self._postRun(config)

    def _preRun(self, config):
        self.agent.syncCounter()
        self.memory = []
        # self.config = config
        # self.n = 0

    def _run(self, config):
        self.runIteration(config)

    def _postRun(self, config):
        return self.memory
    
    # def processMemory(self, memory=None):
    #     memory = parameter(memory, self.memory)
    #     self.agent.addMemory(memory, self.config)

    def runIteration(self, config):
        self.agent.environment.checkTerminated()
        self.agent.syncCounter()
        self.agent.environment.setupIteration(config)
        self._runIteration(config)

    def _runIteration(self, config):
        pass

    def reachGoalContext(self, config):
        if not config.goalContext or not config.changeContext:
            config.result.reachedContextStatus = 'not required'
            return True

        probFirstPass = linearValue(1., 0., (config.goalContext.space.number - 20) / 50)
        if not config.exploitation and random.uniform(0, 1) < probFirstPass:
            config.result.reachedContextStatus = f'aborted (prob={probFirstPass:.3f})'
            return True

        settings = config.plannerSettings.clone(context=True)
        if config.goal and not config.goal.space.matches(config.goalContext.space, kindSensitive=False):
            settings.dontMoveSpaces.append(config.goal.space)

        try:
            path, _, _ = self.planner.planDistance(config.goalContext.setRelative(False), settings=settings)
        except ActionNotFound:
            path = None

        if not path:
            config.result.reachedContextStatus = 'planning failed'
            self.logger.warning(f"Planning failed for pre-goal context {config.goalContext}")
            return False

        self.testPath(path, config.clone(model=None))
        config.result.reachedContext = self.agent.environment.state().context().projection(config.goalContext.space)
        config.result.reachedContextDistance = (config.result.reachedContext - config.goalContext).norm()
        model = path.model
        if not model or config.result.reachedContextDistance >= model.outcomeSpace.maxDistance * self.DISTANCE_MAX_CONTEXT:
            config.result.reachedContextStatus = 'too far from context goal'
            return False

        config.result.reachedContextStatus = 'context successfully reached'
        return True

        # config.goalContext = config.goalContext.convertTo(
        #     self.agent.dataset, kind=SpaceKind.BASIC)

        # try:
        #     paths, _ = self.planner.planDistance(config.goalContext)
        #     self.logger.debug("Planning generated for context goal {} in {} steps"
        #                       .format(config.goalContext, len(paths)), 'PLAN')
        #     self.testGoal(config.goalContext, paths,
        #                   config.clone(model=None))
        # except ActionNotFound:
        #     self.logger.warning("Context planning failed for goal {}, switching to random"
        #                         .format(config.goalContext), 'PLAN')
        #     return False

    def testActions(self, actions, config=MoveConfig()):
        try:
            self.testPath(self.planner.planActions(actions), config)
        except ActionNotFound:
            return

    # def testGoal(self, goal, paths=None, config=MoveConfig()):
    #     results = self.performer.performGoal(goal, paths, config)
    #     # print('Tested', results)
    #     # self.n += InteractionEvent.incrementList(results, self.n)
    #     self.memory += results

    def testPath(self, path, config=MoveConfig()):
        """Test a specific complex action and store consequences in memory."""
        results = self.performer.perform(path, config)
        # self.n += InteractionEvent.incrementList(results, self.n)
        self.memory += results
        # self.n += 1

    def __deepcopy__(self, a):
        newone = type(self).__new__(type(self))
        newone.__dict__.update(self.__dict__)
        newone.parent = None
        newone.modules = []
        newone.agent = None
        newone.__dict__ = copy.deepcopy(newone.__dict__)
        return newone



import pygame
import pymunk
from pymunk import Vec2d

from PIL import Image

from .entity import ProxyEntity
from .entity_manager import EntityManager
from .property import FunctionObservable


class MapProxyEntity(ProxyEntity):
    # def __init__(self, entity, manager=None):
    #     super().__init__(entity, manager=manager)

    # Handlers
    def addPropertyHandler(self, name, handler):
        property_ = self.propertyItem(name)
        # self.handlers[property_] = handler
        self.manager.addPropertyHandler(property_, handler)


class FeatureMap(EntityManager):
    def __init__(self, proxyCls=ProxyEntity):
        EntityManager.__init__(self, 'featureMap', proxyCls=proxyCls)

        self.handlers = {}
        # self.rawHandlers = {}
        self.generators = []
        self.rawGenerators = []
    
    def __repr__(self):
        return f'{self.__class__.__name__} (adding {", ".join([f"{g[0]}.{g[1]}" for g in self.generators])})'

    def addPropertyHandler(self, property_, handler):
        if property_:
            self.handlers[property_] = handler
    
    def addPropertyGenerator(self, entityFilter, name, function, **kwargs):
        self.generators.append((entityFilter, name, function, kwargs))
    
    def customEntityConversion(self, entity):
        for entityFilter, name, function, kwargs in self.generators:
            if entity in self.world.cascadingChildren(entityFilter):
                property_ = FunctionObservable(entity, name, function, **kwargs, proxySpace=True)
                self.rawGenerators.append(property_)
        # for name, handler in self.handlers:
        #     property_ = entity.propertyItem(name)
        #     if property_:
        #         self.rawHandlers[property_] = lambda value: handler(entity, value)
        return entity

    def populateFrom(self, manager):
        for entity in manager.entities():
            entity.convertTo(self)
        self.world.activate()
    
    def setToValues(self, values):
        for value in values:
            handler = self.handlers.get(value.space.boundProperty)
            if handler:
                handler(value.plain())
    
    def updateValues(self, values, dataset):
        for property_ in self.rawGenerators:
            found = False
            space = property_.space
            newValue = property_.observe()
            if dataset:
                newValue = newValue.convertTo(dataset)
            for i, value in enumerate(values):
                if value.space.linkedTo(space):
                    values[i] = newValue
                    found = True
                    break
            if not found:
                values.append(newValue)

    def update(self, values, dataset=None):
        self.setToValues(values)
        self.updateValues(values, dataset)
        return values

    def image(self):
        pass



from exlab.interface.serializer import Serializable

from dinos.data.data import Data, Action, Observation


# Episode -> list of InteractionEvents
class InteractionEvent(Serializable):
    """
    Represents an Event, composed by different actions and outcomes
    :param actions: an ActionList object
    :param outcomes: an Observation object
    """

    def __init__(self, iteration, actions=Data(), primitiveActions=Data(), outcomes=Data(), context=Data()):
        self.iteration = iteration

        self.actions = actions
        self.primitiveActions = primitiveActions
        self.outcomes = outcomes
        self.context = context

        self.actionsRegister = []
        self.primitiveActionsRegister = []
        self.outcomesRegister = []
        self.contextRegister = []

        # Check if action is also present in the outcomes
        actions = self.actions.flat()
        onlyOutcomes = self.outcomes.flat()
        for outcome in list(onlyOutcomes):
            for action in actions:
                if outcome.space.matches(action.space):
                    action.value = outcome.value
                    onlyOutcomes.remove(outcome)

        # Check for no parameter actions (and set it to 1)
        # for action in actions:
        #     if len(action.value) == 0:
        #         action.value = [1]
        self.actions = Action(*actions)
        self.onlyOutcomes = Observation(*onlyOutcomes)
        self.allActions = None

    def _serialize(self, serializer):
        dict_ = serializer.serialize(self, ['actions', 'primitiveActions', 'outcomes', 'context',
                                            'actionsRegister', 'primitiveActionsRegister', 'outcomesRegister',
                                            'contextRegister'])
        return dict_

    # @classmethod
    # def deserialize(cls, dict_, dataset, spaces, loadResults=True):
    #     a = [next(i for i in spaces if i.name == name or i.id == name) for name in dict_['actions']]
    #     y = [next(i for i in spaces if i.name == name or i.id == name) for name in dict_['outcomes']]
    #     c = [next(i for i in spaces if i.name == name or i.id == name) for name in dict_.get('context', [])]
    #     obj = cls(dataset, a, y, c)
    #     return obj

    def clone(self):
        return self.__class__(self.iteration,
                              self.actions.clone(),
                              self.primitiveActions.clone(),
                              self.outcomes.clone(),
                              self.context.clone())

    '''def flatten(self):
        return (self.action.flatten(), self.outcomes.flatten())

    @staticmethod
    def from_flat(data):
        return InteractionEvent(*data)'''

    def incrementIteration(self, n):
        self.iteration += n

    def convertTo(self, spaceManager=None, kind=None, toData=None):
        self.actions = self.actions.convertTo(
            spaceManager=spaceManager, kind=kind, toData=toData)
        self.primitiveActions = self.primitiveActions.convertTo(
            spaceManager=spaceManager, kind=kind, toData=toData)
        self.outcomes = self.outcomes.convertTo(
            spaceManager=spaceManager, kind=kind, toData=toData)
        self.onlyOutcomes = self.onlyOutcomes.convertTo(
            spaceManager=spaceManager, kind=kind, toData=toData)
        self.context = self.context.convertTo(
            spaceManager=spaceManager, kind=kind, toData=toData)
        
        allActions = self.actions.flat() + self.primitiveActions.flat() + self.onlyOutcomes.flat()
        self.allActions = Action(*allActions)

    def addToSpaces(self, cost=1.0):
        for point in self.actions.flat():
            idAction = point.space.addPoint(point, self.iteration, action=True)
            self.actionsRegister.append((point.space.id, idAction))

        if not self.actions == self.primitiveActions:
            for point in self.primitiveActions.flat():
                idAction = point.space.addPoint(point, self.iteration, action=True)
                self.primitiveActionsRegister.append((point.space.id, idAction))

        for point in self.onlyOutcomes.flat():
            idOutcome = point.space.addPoint(point, self.iteration, cost)
            self.outcomesRegister.append((point.space.id, idOutcome))

        if self.context:
            for point in self.context.flat():
                idContext = point.space.addPoint(point, self.iteration, cost)
                self.contextRegister.append((point.space.id, idContext))

    @staticmethod
    def incrementList(list_, currentIteration):
        if not list_:
            return 0
        n = max([event.iteration for event in list_]) + 1
        for event in list_:
            event.incrementIteration(currentIteration)
        return n

    def __repr__(self):
        return f'{self.__class__.__name__}#{self.iteration}({self.actions}, {self.outcomes}, {self.context})'



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import math
import copy
import random
import pickle

from exlab.utils.text import strtab
from exlab.interface.graph import Graph
from exlab.interface.serializer import Serializable

from .space import Space, SpaceKind
from .dataspace import DataSpace
# from ..utils.logging import DataEventHistory, DataEventKind
# from ..utils.io import strtab, getVisual, plotData, visualize
# from dinos.utils.maths import uniformSampling, first
from .data import Data


class Point(Serializable):
    MAX_VALUES = 10
    WINDOW_SIZE = 5
    NEAR_CONTEXT_DISTANCE = 0.04
    ZERO_RANGE = 0.001

    def __init__(self, id_, position, context, positionContext, value=None):
        self.id = id_
        self.position = position
        self.context = context
        self.positionContext = positionContext

        self.zero = position.norm() < self.ZERO_RANGE

        self.values = []
        self.progresses = []
        self.evaluation = 0.
        self.cost = 1.

        if value is not None:
            self.addValue(value)

    def __repr__(self):
        return f'Point #{self.id} {self.position} {self.evaluation}'
    
    def _serialize(self, serializer):
        dict_ = serializer.serialize(
            self, ['id', 'position', 'context', 'positionContext', 'values', 'progresses', 'evaluation', 'cost'])
        return dict_

    @classmethod
    def _deserialize(cls, dict_, serializer, obj=None):
        if obj is None:
            obj = cls(dict_.get('id'),
                      dict_.get('position'),
                      dict_.get('context'),
                      dict_.get('positionContext'),
                      serializer.get('.region.parent'))
        return super()._deserialize(dict_, serializer, obj)

    def _postDeserialize(self, dict_, serializer):
        super()._postDeserialize(dict_, serializer)
        for attr in ['values', 'progresses', 'evaluation', 'cost']:
            setattr(self, attr, dict_.get(attr))
    
    @property
    def lastValue(self):
        return self.values[-1]
    
    def needUpdate(self):
        if len(self.progresses) < 4:
            return True
        if random.uniform(0, 1) < self.evaluation * 10:
            return True
        return False

    def addValue(self, value):
        self.values.append(value)
        self.values = self.values[-self.MAX_VALUES:]
        self.computeEvaluation(self.values, self.WINDOW_SIZE, cost=self.cost)
    
    def computeEvaluation(self, values, windowSize, cost=1.):
        """Compute evaluation of the region for given strategy."""
        if len(values) >= windowSize:
            self.progresses.append(self.meanProgress(values))
            self.progresses = self.progresses[-windowSize:]

        if len(self.progresses) >= 2:
            self.evaluation = np.abs((self.progresses[-1] - self.progresses[0]) / cost)
        elif values:
            self.evaluation = values[-1] / cost
        else:
            self.evaluation = 0.

    def meanProgress(self, values):
        """Compute mean progress according to the evaluation window."""
        return np.mean(values[-self.WINDOW_SIZE:])
    
    def nearContext(self, projectedContext=None):
        if not projectedContext:
            return True
        distance = self.context.distanceTo(projectedContext) / self.context.space.maxDistance
        return distance < self.NEAR_CONTEXT_DISTANCE



class SpaceRegion(Serializable):
    """Implements an evaluation region."""

    REGION_BASED = 0
    REGION_UPDATABLE = 1
    POINT_BASED = 2

    def __init__(self, space, options, bounds=None, parent=None, manager=None, tag='',
                 contextSpace=None, regions=None):
        """
        bounds float list dict: contains min and max boundaries of the region
        options dict: different options used by the evaluation model
        """
        self.explorableSpace = space
        self.explorableContextSpace = space.spaceManager.multiColSpace([space, contextSpace], weight=0.5)
        self.contextSpace = contextSpace

        self.parent = parent
        self._manager = manager

        self.bounds = copy.deepcopy(
            bounds) if bounds else Space.infiniteBounds(self.explorableContextSpace.dim)
        assert(parent is not None or manager is not None)
        self.tag = tag

        self.colsExplorable = self.explorableContextSpace.columnsFor(self.explorableSpace)
        self.colsContext = self.explorableContextSpace.columnsFor(self.contextSpace)

        # Contains the following keys:
        #   'minSurface': the minimum surface of a region to be still splittable
        #   'maxPoints': the maximum of points contained in a region (used to split)
        #   'window': the number of last progress measures to consider for computing region evaluation (to take newer
        #             points only)
        #   'cost': cost of strategy (usually strategies involving a teacher have a higher cost)
        self.options = {
            "cost": 1.0,
            "window": 10,
            "maxAttempts": 20,
            "maxPoints": 40,
            "minSurface": 0.05,
            "pointNumberSplit": 50
        }
        self.options.update(options)

        self.number = 0
        self.points = []
        self.nonZeroPoints = []
        self.progresses = []
        self.evaluation = 0.
        # self.base = Point(None, None, None, None)

        self.leftChild = None
        self.rightChild = None
        self.regions = regions if regions else [self]

        self.splitValue = 0.
        self.splitDim = None

        self.setSplittable()

        if not parent:
            # register history only for the root region
            # self.history = DataEventHistory()
            self.childrenNumber = 0
        else:
            self.lid = self.root().childrenNumber
            self.root().childrenNumber += 1

        self.explorablePreSpace = self.explorableSpace.convertTo(self.dataset, kind=SpaceKind.PRE)

    def __repr__(self):
        cut = strtab(f'{self.splitDim}th d: {self.splitValue:.4f}')
        return f'Region {self.explorableContextSpace} {self.evaluation:.4f}\n    Left: {self.leftChild is not None}\n    <Cut {cut}>\n    Right: {self.rightChild is not None}'

    def _serialize(self, serializer):
        dict_ = serializer.serialize(
            self, ['explorableSpace', 'contextSpace', 'bounds', 'points', 'number', 'progresses', 'evaluation',
                   'leftChild', 'rightChild', 'splitValue', 'splitDim', 'tag'])
        return dict_

    @classmethod
    def _deserialize(cls, dict_, serializer, obj=None):
        if obj is None:
            # leftChild = serializer.deserialize(dict_.get('leftChild'))
            # rightChild = serializer.deserialize(dict_.get('rightChild'))
            obj = cls(serializer.deserialize(dict_.get('explorableSpace')),
                      options=dict_.get('options', {}),
                      bounds=dict_.get('bounds'),
                      parent=serializer.get('.region.parent'),
                      manager=serializer.get('.region.manager'),
                      tag=dict_.get('tag', ''),
                      contextSpace=serializer.deserialize(dict_.get('contextSpace')))
        return super()._deserialize(dict_, serializer, obj)

    def _postDeserialize(self, dict_, serializer):
        super()._postDeserialize(dict_, serializer)
        serializer = serializer.clone(values={
            '.region.parent': self,
            '.region.manager': self.manager,
        })

        for attr in ['number', 'progresses', 'evaluation']:
            setattr(self, attr, dict_.get(attr))
        for attr in ['points']:
            setattr(self, attr, serializer.deserialize(dict_.get(attr)))
        self.nonZeroPoints = [p for p in self.points if not p.zero]

        self.leftChild = serializer.deserialize(dict_.get('leftChild'))
        self.rightChild = serializer.deserialize(dict_.get('rightChild'))

    def root(self):
        root = self
        while root.parent is not None:
            root = root.parent
        return root
    
    @property
    def manager(self):
        return self.root()._manager
    
    @property
    def dataset(self):
        return self.root()._manager.dataset
    
    @property
    def method(self):
        return self.root()._manager.method
    
    def positions(self, withContext=True):
        if withContext:
            return np.array([p.positionContext for p in self.points])
        return np.array([p.position for p in self.points])

    def finiteBounds(self):
        # points = np.array(self.root().getSplitData(withRegions=False)[0])
        positions = self.positions()
        return [(np.min(positions[:, j]) if bmin == -math.inf else bmin,
                 np.max(positions[:, j]) if bmax == math.inf else bmax)
                for j, (bmin, bmax) in enumerate(self.bounds)]

    def maxDistance(self):
        positions = self.positions()
        bounds = np.max(positions, axis=0) - np.min(positions, axis=0)
        return np.sum(bounds ** 2) ** 0.5

    @property
    def splitten(self):
        return self.leftChild is not None

    # def nearestContext(self, context):
    #     self.contextSpace._validate()
    #     context = context.convertTo(kind=SpaceKind.PRE).projection(self.contextSpace).plain()

    #     indices, distances = DataSpace.nearestFromData(np.array(self.points)[:, self.colsContext], context,
    #                                                             n=self.options['maxPoints']//2)

    #     return indices, distances

    def nearContext(self, context, tolerance=0.01):
        if not self.contextSpace:
            return True

        self.contextSpace._validate()
        context = context.convertTo(kind=SpaceKind.PRE).projection(self.contextSpace).plain()
        tolerance = self.contextSpace.maxDistance * tolerance

        # Context in bounds
        inBounds = True
        for i, col in enumerate(self.colsContext):
            if context[i] < self.bounds[col][0] - tolerance or context[i] > self.bounds[col][1] + tolerance:
                inBounds = False
                break
        return inBounds

    def controllableContext(self):
        if not self.contextSpace:
            return False
        return self.dataset.controllableSpaces(self.contextSpace)
    
    def updatePoints(self, aroundPoint, closeIds=None):
        for point in self.nonZeroPoints:
            if point.id in closeIds and point != aroundPoint:
                self.updatePoint(point)

    def updatePoint(self, point):
        pass

    def computeEvaluation(self):
        if self.method == self.POINT_BASED:
            self.evaluation = max(p.evaluation for p in self.points)
        elif self.method == self.REGION_BASED:
            values = [p.lastValue for p in self.points]
            Point.computeEvaluation(self, values, self.options['window'])

    def findRandom(self, allowZeros=False):
        if allowZeros:
            return random.choice(self.points)
        if not self.nonZeroPoints:
            return None
        return random.choice(self.nonZeroPoints)
 
    def findBest(self, context=None, changeContext=True, favoriseBest=1.):
        dataChangeContext, dataCurrentContext = [], []
        if context:
            context = context.projection(self.contextSpace)
            dataCurrentContext += [point for point in self.nonZeroPoints if point.nearContext(context)]

        if changeContext:
            dataChangeContext += list(self.nonZeroPoints)
            changeContext = self.chooseToChangeContext(max(x.evaluation for x in dataChangeContext) if dataChangeContext else -math.inf,
                                                       max(x.evaluation for x in dataCurrentContext) if dataCurrentContext else -math.inf)

        data = dataChangeContext if changeContext else dataCurrentContext
        # data.sort(key=lambda x: -x[1])

        if not data:
            return self.findRandom(), changeContext
        return random.choices(data, weights=[x.evaluation for x in data])[0], changeContext
    
    def createRandomPoint(self):
        return self.explorableSpace.point(np.array([random.uniform(bmin, bmax) for bmin, bmax in self.finiteBounds()])[self.colsExplorable])
    
    def chooseToChangeContext(self, bestChangedContext, bestCurrentContext):
        if bestCurrentContext > bestChangedContext * 4:
            return False
        if bestChangedContext > bestCurrentContext * 4:
            return True
        return random.uniform(0, bestChangedContext + bestCurrentContext) < bestChangedContext
    
    def getGoalContext(self, point):
        if self.controllableContext():
            return point.context
        return self.explorableSpace.point(self.explorablePreSpace.getData([point.id])[0]).setRelative(False)

    def addPoint(self, id_, point, context, value, closeIds=None, firstAlwaysNull=True, populating=False):
        """Add a point and its value in the attached evaluation region."""
        pointContext = Data.npPlainData(point.extends(context), self.explorableContextSpace)
        # point = Data.plainData(point, self.explorableSpace)
        # context = Data.plainData(context, self.contextSpace)
        point = point.projection(self.explorableSpace, kindSensitive=True)
        context = context.projection(self.contextSpace, kindSensitive=True)
        assert len(pointContext) == len(self.bounds)

        # print("ADDING 1 POINT", point)
        # if filterNull and np.all(np.array(point)[self.colsTarget]==0):
        #     print("NULL POINT")
        #     return

        # if firstAlwaysNull and len(self.points) == 0:
        #     value = 0.

        p = Point(id_, point, context, pointContext, value)
        self._addPointData(p, closeIds=closeIds, firstAlwaysNull=firstAlwaysNull, populating=populating)

    def _addPointData(self, point, closeIds=None, firstAlwaysNull=True, populating=False):
        self.points.append(point)
        if not point.zero:
            self.nonZeroPoints.append(point)
        self.number += 1

        if not populating and self.method == self.POINT_BASED:
            self.updatePoints(point, closeIds)
        self.computeEvaluation()

        # addedInChildren = False
        if not self.splitten and self.splittable and self.number > self.options['pointNumberSplit']:
            # Leave reached, region must be splitten
            self.randomCut(self.options['maxAttempts'])
            self.split()
        elif self.splitten:
            # Traverse tree
            # addedInChildren = True
            child = self.leftChild if point.positionContext[self.splitDim] < self.splitValue else self.rightChild
            child._addPointData(point, closeIds=closeIds, populating=populating)

        # Remove oldest points and pointValues
        if self.method == self.POINT_BASED:
            pass
        else:
            self.points = self.points[-self.options['maxPoints']:]

        # if not addedInChildren and not populating and self.manager:
        #     self.manager.logger.debug(f'Adding point [{", ".join(["{:.4f}".format(p) for p in point])}] with value {value:.3e} to region {self}', tag=self.tag)

    # Splitting process
    def setSplittable(self):  # Change it to use max depth tree instead ???
        """Check if the region is splittable according to its surface."""
        surface = 1.
        for bmin, bmax in self.bounds:
            surface *= (bmax - bmin)
        self.splittable = (surface > self.options['minSurface'])

    def split(self):
        """Split region according to the cut decided beforehand."""

        if self.splitDim is not None:
            return
        if self.manager:
            self.manager.logger.debug(f'Splitting along dim {self.splitDim}: {self.splitValue:.4f} for {self}', tag=self.tag)

        # Create child regions boundaries
        leftBounds = copy.deepcopy(self.bounds)
        leftBounds[self.splitDim][1] = self.splitValue
        rightBounds = copy.deepcopy(self.bounds)
        rightBounds[self.splitDim][0] = self.splitValue

        # Create empty child regions
        self.leftChild = self.__class__(self.explorableSpace, self.options, bounds=leftBounds,
                                        parent=self, contextSpace=self.contextSpace, regions=self.regions)
        self.rightChild = self.__class__(self.explorableSpace, self.options, bounds=rightBounds,
                                         parent=self, contextSpace=self.contextSpace, regions=self.regions)
        # root = self.root()
        # root.history.append(root.manager.getIteration(), DataEventKind.ADD, [(str(self.leftChild.lid), self.leftChild.serialize())])
        # root.history.append(root.manager.getIteration(), DataEventKind.ADD, [(str(self.rightChild.lid), self.rightChild.serialize())])

        # Add them in the list of regions (that's the only reason we need access to the list of all regions!!!)
        self.regions.append(self.leftChild)
        self.regions.append(self.rightChild)

        #print("-----------------------------")
        #print("Split done")
        #print("Region [" + str(self.bounds['min']) + ", " + str(self.bounds['max']) + "] ----> Left ["
        #      + str(left_region.bounds['min']) + ", " + str(left_region.bounds['max']) + "] (")

        #left = 0
        #right = 0
        #for s in range(len(self.points)):
        #    for i in range(len(self.points[s])):
        #        point = self.points[s][i]
        #        if point[self.splitDim] < self.splitValue:
        #            left += 1
        #        else:
        #            right += 1

        #print("Split put " + str(left) + " points left, " + str(right) + " points right.")
        #print("Split done on dimension " + str(self.splitDim) + " at value: " + str(self.splitValue))

        # Add all points of the parent region in the child regions according to the cut
        for point in self.points:
            child = self.leftChild if point.positionContext[self.splitDim] < self.splitValue else self.rightChild
            child._addPointData(point, populating=True)

    # def greedyCut(self):
    #     """UNTESTED method to define a cut greedily."""
    #     i = range(len(self.points))
    #     maxQ = 0.

    #     # For each dimension
    #     for d in range(len(self.bounds['min'])):
    #         # Sort the points in the dimension
    #         i.sort(key=lambda j: self.points[j][d])
    #         for p in range(len(i)-1):
    #             progress_left = 0.
    #             progress_right = 0.

    #             left = copy.deepcopy(i[0:(p+1)])
    #             right = copy.deepcopy(i[(p+1):len(i)])

    #             n = min(len(left), self.options['window'])
    #             if n < len(left):
    #                 left.sort()
    #             for k in range(n):
    #                 progress_left += self.pointValues[left[len(left)-k-1]]
    #             progress_left /= float(n)

    #             n = min(len(right), self.options['window'])
    #             if n < len(right):
    #                 right.sort()
    #             for k in range(n):
    #                 progress_right += self.pointValues[right[len(right)-k-1]]
    #             progress_right /= float(n)

    #             delta_p = (progress_right - progress_left)**2
    #             Q = delta_p * len(left) * len(right)

    #             if Q > maxQ:
    #                 # Choose mean between two points as the cut
    #                 self.splitDim = d
    #                 self.splitValue = (
    #                     self.points[i[p]][d] + self.points[i[p+1]][d]) / 2.

    # Careful !!! The tree can't handle it if it has only one value multiple times !!!
    def randomCut(self, numberAttempts):
        """Define the cut by testing a few cuts per dimension."""
        if self.manager:
            self.manager.logger.debug2(f'Trying to cut region {self}', tag=self.tag)
        i = list(range(len(self.points)))
        #n = int(math.ceil(len(self.points)/(numberAttempts+1)))
        n = float(len(self.points)) / float(numberAttempts + 1)
        maxQ = 0.
        #max_card = 0
        #splitValue_card = 0.
        #splitDim_card = -1

        # For each dimension
        #from pprint import pprint
        for d in range(len(self.bounds)):
            i.sort(key=lambda j: self.points[j].positionContext[d])
            # For each attempt
            for k in range(numberAttempts):
                progress_left = 0.
                progress_right = 0.

                # Sort points by age
                #left = copy.deepcopy(i[0:((k+1)*n)])
                #right = copy.deepcopy(i[((k+1)*n):len(i)])
                #left.sort()
                #right.sort()

                # Id of the item following cut
                idcut = int(math.floor((k+1)*n))

                splitValue = (self.points[i[idcut-1]].positionContext[d] + self.points[i[idcut]].positionContext[d]) / 2.

                # Make sure we are not trying to split something unsplittable
                if self.points[i[idcut-1]].positionContext[d] == splitValue:
                    continue

                left = []
                right = []
                for j in range(len(self.points)):
                    if self.points[j].positionContext[d] < splitValue:
                        left.append(j)
                    else:
                        right.append(j)

                # Retain only the points inside evaluation window
                left_filtered = left[(
                    len(left)-min(len(left), self.options['window'])):len(left)]
                right_filtered = right[(
                    len(right)-min(len(right), self.options['window'])):len(right)]

                # Compute progress for each part
                for j in left_filtered:
                    progress_left += self.points[j].lastValue
                progress_left /= float(max(len(left_filtered), 1))
                for j in right_filtered:
                    progress_right += self.points[j].lastValue
                progress_right /= float(max(len(right_filtered), 1))

                # Compute delta_p and Q
                delta_p = (progress_left - progress_right)**2
                card = len(left) * len(right)
                if len(left) > 0 and len(right) > 0:
                    Q = delta_p * float(card)
                else:
                    Q = -float('inf')

                if Q > maxQ:
                    # Choose mean between two points as the cut
                    self.splitDim = d
                    self.splitValue = splitValue
                    maxQ = Q
        if self.manager:
            self.manager.logger.debug2(f'Found split along dim {self.splitDim}: {self.splitValue:.4f} for {self}', tag=self.tag)

    def getSplitData(self, withRegions=True):
        points = self.positions()
        pointValues = [point.value for point in self.points]

        regions = []
        if withRegions and not self.splitten:
            regions.append((self.evaluation, self.finiteBounds()))

        for child in [self.leftChild, self.rightChild]:
            if child:
                p, pv, r = child.getSplitData(withRegions=withRegions)
                points += p
                pointValues += pv
                if withRegions:
                    regions += r
        return points, pointValues, regions

    def _cuts(self):
        if not self.splitten:
            return []
        return [(self.splitDim, self.splitValue)] + self.leftChild._cuts() + self.rightChild._cuts()

    def cuts(self):
        _cuts = self._cuts()
        return _cuts

    # Visual
    def visualizeData(self, options={}, outcomeOnly=True, contextOnly=False, absoluteProgress=False):
        g = Graph(title=f'Regions from {self}', options=options)
        points, pointValues, regions = self.getSplitData()

        points = np.array(points)
        pointValues = np.clip(np.array(pointValues), -100, 100)

        if absoluteProgress:
            pointValues = np.abs(pointValues)

        # Filter
        if contextOnly:
            cols = self.colsContext
            points = points[:, cols]
        elif outcomeOnly:
            cols = self.colsExplorable
            points = points[:, cols]
        else:
            cols = np.arange(self.explorableContextSpace.dim)

        # pvMinimum = np.min(pointValues)
        # pvMaximum = np.max(pointValues)
        evaluation = np.array([r[0] for r in regions])
        evalMinimum = np.min(evaluation)
        # evalMaximum = np.max(evaluation)
        evaluation = (evaluation - evalMinimum) / max(0.001, np.max(evaluation) - evalMinimum)

        for region in regions:
            bounds = np.array(region[1])[cols]
            alpha = (0.2 + region[0]) / 1.2
            if len(bounds) == 2:
                g.rectangle((bounds[0][0], bounds[1][0]), bounds[0][1] - bounds[0]
                            [0], bounds[1][1] - bounds[1][0], alpha=alpha, zorder=-1, border=True)
            else:
                g.rectangle((bounds[0][0], -1), bounds[0][1] -
                            bounds[0][0], 2, alpha=alpha, zorder=-1, border=True)
        g.scatter(points, color=pointValues, colorbar=True)
        return g

    # def getRegionsVisualizer(self, prefix='', outcomeOnly=True, contextOnly=False, absoluteProgress=False):
    #     """Return a dictionary used to visualize evaluation regions."""
    #     points, pointValues, regions = self.getSplitData()

    #     points = np.array(points)
    #     pointValues = np.clip(np.array(pointValues), -100, 100)
    #     if absoluteProgress:
    #         pointValues = np.abs(pointValues)
    #     # regions = np.array(regions)
    #     if contextOnly:
    #         cols = self.colsContext
    #         points = points[:, cols]
    #     elif outcomeOnly:
    #         cols = self.colsTarget
    #         points = points[:, cols]
    #     else:
    #         cols = np.arange(self.space.dim)

    #     pvMinimum = np.min(pointValues)
    #     pvMaximum = np.max(pointValues)
    #     # pointValues = (pointValues - pvMinimum) / max(0.001, pvMaximum - pvMinimum)
    #     # print(pointValues)

    #     evaluation = np.array([r[0] for r in regions])
    #     evalMinimum = np.min(evaluation)
    #     evalMaximum = np.max(evaluation)
    #     evaluation = (evaluation - evalMinimum) / \
    #         max(0.001, evalMaximum-evalMinimum)

    #     import matplotlib.patches as patches

    #     def plotInterest(region, ax, options):
    #         bounds = np.array(region[1])[cols]
    #         alpha = (1. + region[0]) / 2.
    #         if len(bounds) == 2:
    #             ax.add_patch(patches.Rectangle((bounds[0][0], bounds[1][0]), bounds[0][1] - bounds[0][0],
    #                                            bounds[1][1] - bounds[1][0], alpha=alpha, zorder=-1))
    #             ax.add_patch(patches.Rectangle((bounds[0][0], bounds[1][0]), bounds[0][1] - bounds[0][0],
    #                                            bounds[1][1] - bounds[1][0], alpha=alpha, zorder=-1, fill=False))
    #         else:
    #             ax.add_patch(patches.Rectangle(
    #                 (bounds[0][0], -1), bounds[0][1] - bounds[0][0], 2, alpha=alpha, zorder=-1))

    #     def lambdaInit(region):
    #         return lambda fig, ax, options: plotInterest(region, ax, options)

    #     bounds = np.array(self.finiteBounds())[cols]
    #     return getVisual(
    #         [lambdaInit(region) for region in regions] +
    #         [lambda fig, ax, options: plotData(
    #             points, fig, ax, options, color=pointValues, colorbar=True)],
    #         minimum=bounds[:, 0].tolist(),
    #         maximum=bounds[:, 1].tolist(),
    #         title='{}$Interest \in [{:.2f}, {:.2f}]$, $Progress \in [{:.2f}, {:.2f}]$\n{} points reached, {} regions'.format(prefix, evalMinimum,
    #                                                                                                                          evalMaximum, pvMinimum,
    #                                                                                                                          pvMaximum, points.shape[0], len(regions))
    #     )

    # Plot
    # def plot(self, outcomeOnly=True, contextOnly=False, absoluteProgress=False):
    #     visualize(self.getRegionsVisualizer(outcomeOnly=outcomeOnly,
    #                                         contextOnly=contextOnly, absoluteProgress=absoluteProgress))

    # *** Deprecated ***

    # def plot2(self, fig_id, color, norm):
    #     """
    #     Plot evaluation region as a rectangle patch which transparency indicates evaluation.

    #     fig_id int
    #     color string: indicates the colour of the patch
    #     norm float: used to normalize the patches transparency
    #     """
    #     num_plots = min(len(self.evaluation), 10)
    #     plt.ion()
    #     fig = plt.figure(fig_id)
    #     plt.show()
    #     if norm == 0.0:
    #         norm2 = 1.0
    #     else:
    #         norm2 = norm
    #     n = (num_plots + 1) / 2
    #     for s in range(num_plots):
    #         ax = fig.add_subplot(str(n) + "2" + str(s+1))
    #         if len(self.bounds['min']) == 1:
    #             p = patches.Rectangle(
    #                 (self.bounds['min'][0], -0.5), (self.bounds['max']
    #                                                 [0]-self.bounds['min'][0]), 1,
    #                 facecolor=color, alpha=-self.evaluation[s]/norm2)
    #             ax.add_patch(p)
    #         elif len(self.bounds['min']) == 2:
    #             p = patches.Rectangle(
    #                 (self.bounds['min'][0], self.bounds['min'][1]
    #                  ), (self.bounds['max'][0]-self.bounds['min'][0]),
    #                 (self.bounds['max'][1]-self.bounds['min'][1]),
    #                 facecolor=color, alpha=-self.evaluation[s]/norm2)
    #             ax.add_patch(p)
    #     plt.draw()

    # def plot_v2(self, norm, ax, options):
    #     """Plot evaluation region as a rectangle patch with transparency indicating evaluation."""
    #     num_plots = min(len(self.evaluation), 10)
    #     if norm == 0.0:
    #         norm2 = 1.0
    #     else:
    #         norm2 = norm
    #     n = (num_plots + 1) / 2
    #     if len(self.bounds['min']) == 1:
    #         p = patches.Rectangle(
    #             (self.bounds['min'][0], -0.5), (self.bounds['max']
    #                                             [0]-self.bounds['min'][0]), 1,
    #             facecolor=options['color'], alpha=-self.evaluation[s]/norm2)
    #         ax.add_patch(p)
    #     elif len(self.bounds['min']) == 2:
    #         p = patches.Rectangle(
    #             (self.bounds['min'][0], self.bounds['min'][1]
    #              ), (self.bounds['max'][0]-self.bounds['min'][0]),
    #             (self.bounds['max'][1]-self.bounds['min'][1]),
    #             facecolor=options['color'], alpha=-self.evaluation[s]/norm2)
    #         ax.add_patch(p)

    # Api
    # def apiget(self, range_=(-1, -1)):
    #     if self.parent:
    #         return {}
    #     # , 'evaluation': self.im.get_range(range_)}
    #     return {'regions': self.history.get_range(range_)}



import numpy as np
from pymunk import Vec2d

import random

from dinos.environments.scene import SceneSetup
from dinos.evaluation.tests import UniformGridTest, PointsTest

from panda3d.core import Point2, Point3, Vec3, Vec4
from dinos.environments.engines.tools.panda import PandaTools

from .environment import PlaygroundEnvironment
from .cylinder import Cylinder
from .button import Button
from .agent import Agent
from .wall import Wall


class BaseScene(SceneSetup):
    RESET_ITERATIONS = 10

    # Setup
    def _baseSetup(self):
        self.iterationReset = 0
        self.pandaInit = False

    # Resets
    def countReset(self, forceReset=False):
        self.iterationReset += 1
        if self.iterationReset >= self.RESET_ITERATIONS or forceReset:
            self.iterationReset = 0
            return True
        return False

    def setupIteration(self, config):
        pass

    def setupPreTest(self, test):
        self.reset()

    def _reset(self):
        pass
        # for obj in [self.agent]:
        #     obj.body.position = obj.coordsInit


class EmptyScene(BaseScene):
    DESCRIPTION = 'Just a mobile robot with no obstacle and no other objects.\n' +\
                  'Use: learning how to move.'
    


    # Setup
    def _setup(self):
        self._baseSetup()

        # Add agent
        self._setupAgent()

        # Add cylinders
        # self.world.addChild(Cylinder((400, 500), name='Cylinder3', color=(240, 0, 0), movable=False))
        # self.world.addChild(Cylinder((300, 500), name='Cylinder4', color=(240, 0, 0), movable=False))

        # self.world.addChild(Button((50, 50), name='Button1'))
        # self.world.addChild(Button((500, 500), name='Button2'))

        # self.world.addChild(Cylinder((200, 300), name='Cylinder1'))
        # self.world.addChild(Cylinder((500, 300), name='Cylinder2', color=(128, 224, 0)))
    

        # self.agent = Agent((200, 400), radius=30, name='agent',
        #                    omni=True, xydiscretization=self.world.xydiscretization)
        # self.world.addEntity(self.agent)
    
    def _setupAgent(self):
        self.world.addChild(Agent((300, 300), name='Agent'))

    # Tests
    def _setupTests(self):
        self._testAgentMoving()

    def _testAgentMoving(self):
        points = [(300, 300),
                  (500, 300),
                  (400, 200),
                  (400, 400),
                  (300, 200),
                  (300, 400),
                  (500, 400),
                  (500, 200)]
        test = PointsTest('agent-moving', self.world.cascadingProperty('Agent.position').space, points, relative=False)
        self.addTest(test)

    # Resets
    def setupEpisode(self, config, forceReset=False):
        if self.countReset(forceReset):
            self._resetAgent()
    
    def setupPreTest(self, test):
        self.world.child('Agent').body.position = (400, 300)
    
    def setupPreTestPoint(self, test, point):
        self.world.child('Agent').body.position = (400, 300)
    
    def _resetAgent(self, rand=True):
        self.world.child('Agent').body.position = (random.choice([100, 200, 300, 400]), random.randint(150, 450)) if rand else (300, 300)
    
    def _draw(self, base, drawOptions):
        if drawOptions.panda:
            if not self.pandaInit:
                self.pandaInit = True
            
            W, H = 600, 550
            base.cam.setPos(W * 0.4, H * 1.8, H * 2)
            base.cam.lookAt(W // 2, H // 2, 0)

            pt = PandaTools()
            pt.color = Vec4(0.59, 0.54, 0.16, 1.)
            # pt.horizontalPlane(Point3(0, 0, 0), Point3(W, H, 0))
            pt.horizontalPlane(Point3(50, 50, 0), Point3(550, 500, 0))
            pt.color = Vec4(0.69, 0.7, 0.71, 1.)
            Z = 10000
            pt.horizontalPlane(Point3(-Z, -Z, -1), Point3(Z, Z, -1))
            node = pt.end()

            nodePath = base.render.attachNewNode(node)
            nodePath.setPos(0., 0., 0.)
            nodePath.setShaderAuto()


class RoomWithWallsScene(EmptyScene):
    DESCRIPTION = 'Just a mobile robot with walls and no other objects.\n' +\
                  'Use: learning how to move.'

    # Setup
    def _setup(self):
        self._baseSetup()

        self._setupAgent()
        self._setupWalls()

    def _setupWalls(self):
        self.walls = []

        # outer walls
        outw = 6
        outw2 = outw // 2
        self.walls += [Wall((50.0, 50.0), (550.0, 50.0), outw),
                       Wall((550.0, 50.0 - outw2 + 1), (550.0, 500.0 + outw2), outw),
                       Wall((550.0, 500.0), (50.0, 500.0), outw),
                       Wall((50.0, 500.0 + outw2), (50.0, 50.0 - outw2 + 1), outw)]

        # inner walls
        w = 20
        self.walls += [Wall((260.0, 200.0), (260.0, 400.0), w)]
        # ,Wall((335.0, 50.0), (335.0, 290.0), w)
        # self.walls += [Wall((285.0, 380.0), (315.0, 380.0), w),
        #                Wall((285.0, 410.0), (315.0, 410.0), w)]

        for wall in self.walls:
            self.world.addChild(wall)

    # Tests
    def _setupTests(self):
        self._testAgentMoving()
        self._testAgentMovingObstacles()

    def _testAgentMovingObstacles(self):
        # Test
        # points = [(300, 200),
        #           (300, 400),
        #           (500, 400),
        #           (500, 200)]
        # test = PointsTest('agent-moving', self.world.cascadingProperty('Agent.position').space, points, relative=False)
        # self.addTest(test)
        # EmptyScene._testAgentMoving(self)

        # Test
        points = [(150, 250),
                  (350, 250),
                  (150, 350),
                  (350, 350)]
        test = PointsTest('agent-moving-obstacles', self.world.cascadingProperty('Agent.position').space, points, relative=False)

        def prePoint(point):
            point = point.plain()
            self.world.child('Agent').body.position = (350 if point[0] < 260 else 150,
                                                       250 if point[1] < 300 else 350)

        test.prePoint = prePoint
        self.addTest(test)


class OneCylinderScene(EmptyScene):
    DESCRIPTION = 'Just a mobile robot with no obstacle and no other objects.\n' +\
                  'Use: learning how to move.'

    # Setup
    def _setup(self):
        self._baseSetup()

        self._setupAgent()

        # Add cylinders
        self._setupCylinder1()

        # self.world.addChild(Cylinder((200, 300), name='Cylinder1'))
        # self.world.addChild(Cylinder((500, 300), name='Cylinder2'))
        # self.world.addChild(Cylinder((400, 500), name='Cylinder3', color=(240, 0, 0), movable=False))
        # self.world.addChild(Cylinder((300, 500), name='Cylinder4', color=(240, 0, 0), movable=False))

        # self.world.addChild(Button((50, 50), name='Button1'))
        # self.world.addChild(Button((500, 500), name='Button2'))

        # self.world.addChild(Cylinder((200, 300), name='Cylinder1'))
        # self.world.addChild(Cylinder((500, 300), name='Cylinder2', color=(128, 224, 0)))

        # self.agent = Agent((200, 400), radius=30, name='agent',
        #                    omni=True, xydiscretization=self.world.xydiscretization)
        # self.world.addEntity(self.agent)

    def _setupCylinder1(self):
        self.world.addChild(Cylinder((450, 300), name='Cylinder1'))

    def _setupCylinder2(self):
        self.world.addChild(Cylinder((500, 300), name='Cylinder2'))

    def _setupCylinder3Fixed(self):
        self.world.addChild(Cylinder((400, 500), name='Cylinder3', color=(240, 0, 0), movable=False))

    # Tests
    def _setupTests(self):
        self._testAgentMoving()
        self._testMovingCylinder()

    def _testMovingCylinder(self):
        # Test
        points = [(-25, 0),
                  (-25, 1),
                  (25, 0),
                  (25, 1)]
        test = PointsTest('cylinder1-moving', self.world.cascadingProperty('Agent.position').space, points, relative=True)

        def prePoint(point):
            point = point.plain()
            self.world.child('#Cylinder1').body.position = (150 if point[0] < 0 else 350, 300)
            self.world.child('Agent').body.position = (200 if point[0] < 0 else 300, 150 if point[1] == 1 else 300)

        test.prePoint = prePoint
        self.addTest(test)

    # Resets
    def setupEpisode(self, config, forceReset=False):
        if self.countReset(forceReset):
            self._resetAgent()
            self._resetCylinders()

    def setupPreTest(self, test):
        self._resetAgent()
        self.world.child('#Cylinder1').body.position = (450, 300)
        if self.world.child('#Cylinder2'):
            self.world.child('#Cylinder2').body.position = (300, 150)

    def _resetCylinders(self):
        cyl1 = self.world.child('#Cylinder1').body
        cyl1.position = (random.choice([150, 350]), random.randint(200, 400))

        # pos = self.world.child('Agent').body.position

        # distance = 40. if self.environment.iteration < 100 else 120.

        # obj = self.world.child('#Cylinder1').body
        # if self.world.child('#Cylinder2'):
        #     if random.uniform(0, 1) < 0.5:
        #         obj2 = obj
        #         obj = self.world.child('#Cylinder2').body
        #     else:
        #         obj2 = self.world.child('#Cylinder2').body
        #     obj2.position = pos + Vec2d(240. + np.random.uniform(0.), 0).rotated(np.random.uniform(2*np.pi))
        # obj.position = pos + Vec2d(distance + np.random.uniform(0.), 0).rotated(np.random.uniform(2*np.pi))


class CylinderWallsScene(OneCylinderScene, RoomWithWallsScene):
    # Setup
    def _setup(self):
        self._baseSetup()

        self._setupAgent()
        self._setupWalls()
        self._setupCylinder1()


class OneCylinderPlusFixedScene(OneCylinderScene):
    DESCRIPTION = 'Just a mobile robot with no obstacle and no other objects.\n' +\
                  'Use: learning how to move.'

    # Setup
    def _setup(self):
        self._baseSetup()

        self._setupAgent()

        # Add cylinders
        self._setupCylinder1()
        self._setupCylinder3Fixed()


PlaygroundEnvironment.registerScene(EmptyScene, True)
PlaygroundEnvironment.registerScene(RoomWithWallsScene)
PlaygroundEnvironment.registerScene(OneCylinderScene)
PlaygroundEnvironment.registerScene(OneCylinderPlusFixedScene)




'''
    File name: planner.py
    Author: Alexandre Manoury
    Python Version: 3.6
'''

import numpy as np
from scipy.spatial.distance import euclidean

from collections import namedtuple
import itertools
import random
import math

from .data import Data, Goal, SingleAction, Action, ActionList

"""
Paths[
    Path[
        PathNode(Goal, Action)
    , ...]
, ...]
"""


class ActionNotFound(Exception):
    def __init__(self, message, minDistanceReached=None):
        super().__init__(message)
        self.minDistanceReached = minDistanceReached


# SubGoal = namedtuple('SubGoal', ['goal', 'actions'])


# class ExecutionUnit(object):
#     def __init__(self, ):



# class Paths(object):
#     """
#     Represents multiple Path that should be executed simultaneously
#     """

#     def __init__(self, paths=[]):
#         self.__paths = paths
#         assert(len(paths) == 1, 'Paths object may currently only contain 1 path!')

#     @property
#     def paths(self):
#         return self.__paths

#     @paths.setter
#     def paths(self, paths):
#         self.__paths = paths

#     def length(self):
#         return np.sum([path.length() for path in self.__paths])

#     def extends(self, paths):
#         assert len(self) == 1
#         assert len(paths) == 1
#         self.__paths[0].extends(paths.__paths[0])
#         return self

#     def getGroupedActionList(self):
#         if len(self) > 1:
#             return [(None, self.getActionList())]
#         return self.__paths[0].getGroupedActionList()

#     def getActionList(self):
#         actions = itertools.zip_longest(
#             *[path.getActionList() for path in self.__paths])
#         actions = [Action(*[_f for _f in tup if _f]) for tup in actions]
#         actions_ = []
#         for a in actions:
#             suba = itertools.zip_longest(
#                 *[s if isinstance(s, ActionList) else [s] for s in a])
#             suba = [Action(*[_f for _f in tup if _f]) for tup in suba]
#             suba_ = []
#             for s in suba:
#                 s_ = []
#                 for s2 in s:
#                     if isinstance(s2, ActionList):
#                         s_ += s2.get()
#                     else:
#                         s_.append(s2)
#                 suba_.append(Action(*s_))
#             actions_ += suba_
#         #print("{} --> {}".format(actions, actions_))
#         return ActionList(*actions_)

#     def __iter__(self):
#         return self.__paths.__iter__()

#     def __len__(self):
#         return len(self.__paths)

#     def __getitem__(self, key):
#         return self.__paths[key]

#     def toStr(self, short=False):
#         parts = ' | '.join([path.toStr(short=True) for path in self.__paths])
#         return f'Paths({parts})'

#     def __repr__(self):
#         return self.toStr()


class Path(object):
    """
    Represents multiple goal nodes to be executed in order
    """

    def __init__(self, nodes=[], goal=None, planSettings=None):
        self.__nodes = nodes
        self.goal = goal
        self.planSettings = planSettings  # .clone() if planSettings is not None else None

    @property
    def nodes(self):
        return self.__nodes
    
    @property
    def model(self):
        if len(self.__nodes) > 0:
            return self.__nodes[0].model
        return None

    # def getGroupedActionList(self):
    #     return [node.getGroupedActionList() for node in self]

    # def getActionList(self):
    #     return [action for node in self for action in node.getActionList()]

    def length(self):
        return np.sum([node.length() for node in self.__nodes])
    
    def extends(self, path):
        self.__nodes += path.__nodes

    def __len__(self):
        return len(self.__nodes)

    def __iter__(self):
        return self.__nodes.__iter__()

    def __add__(self, other):
        return self.__class__(self.__nodes + other.__nodes)

    def __getitem__(self, key):
        return self.__nodes[key]
    
    def __bool__(self):
        return len(self) > 0

    def toStr(self, short=False):
        parts = '\n-> '.join([node.toStr(short=True) for node in self.__nodes])
        return f'<<\n   {parts}\n>>'

    def __repr__(self):
        return self.toStr()
    
    # @classmethod
    # def fromActionList(cls, actionList, goal=None):
    #     return cls([PathNode.fromAction(action) for action in actionList], goal=goal)


class PathNode(object):
    """
    Represents a couple (action, goal) where the action should reach the goal
    """

    def __init__(self, action=None, goal=None, model=None, pos=None, absPos=None, parent=None, state=None):
        super().__init__()
        self.action = action
        self.goal = goal

        self.model = model
        self.pos = pos
        self.absPos = absPos
        self.parent = parent
        self.state = state

        self.failures = 0
        self.cantConnectToGoal = False

        self.context = None
        self.execution = None

        self.valid = False
    
    def validate(self):
        self.valid = True
    
    def penalty(self):
        return 1 + 0.1 * self.failures ** 1.1

    def length(self):
        return self.goal.length()
        # if self.paths:
        #     return Paths.length(self)
        # else:
        #     return self.goal.length()
    
    def createPath(self, goal=None, planSettings=None):
        path = []
        node = self
        while node.parent is not None:
            path.append(node)
            node.valid = True
            #dist += euclidean(zeroPlain, node.goal.plain())
            node = node.parent
        return Path(list(reversed(path)), goal=goal, planSettings=planSettings)

    # def getGroupedActionList(self):
    #     return self, self.getActionList()
    #     # action = self.getActionList()
    #     # return (self, action if isinstance(action, ActionList) else ActionList(action))

    # def getActionList(self):
    #     if not self.paths:
    #         return ActionList(self.action)
    #     else:
    #         return super().getActionList()

    def toStr(self, short=False):
        # paths = super().toStr(True) if self.paths else ''
        paths = ''
        return f"{'' if short else 'Node'} {self.action.__class__.__name__}({self.action.toStr(short=True) if self.action else ''} - to reach→ {self.goal.toStr(short=True) if self.goal else 'NoGoal'}){paths}"
        # return "Node: {} ({}, {}) {} [{}]".format(self.goal, self.action, self.model, paths, self.state)

    def __repr__(self):
        return self.toStr()
    
    # @classmethod
    # def fromAction(cls, action):
    #     return cls(action=action)



import setuptools
from distutils.core import setup
from Cython.Build import cythonize
import numpy as np

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="dinos",
    version="0.1.1",
    author="Alexandre Manoury",
    author_email="alex@pika.tf",
    description="Active Learning Environments",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Pikalchemist/dinos",
    ext_modules=cythonize("dinos/data/operations.pyx"),
    include_dirs=[np.get_include()],
    packages=setuptools.find_packages(exclude=('tests', 'examples')),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=["exlab", "Cython", "matplotlib", "numpy", "panda3d", "Pillow", "pygame", "pymunk", "scikit-learn", "scipy", "sklearn"],
)



import numpy as np
import random
import copy
import math

import time

# import matplotlib.pyplot as plt
# from mpl_toolkits.mplot3d import Axes3D

from ..agent import Agent

from exlab.utils.io import parameter
from exlab.interface.graph import Graph

from dinos.utils.move import MoveConfig
# from ...utils.maths import uniformSampling, iterrange

# from ...data.data import InteractionEvent
# from ...data.dataset import Dataset
# from ...models.regression import RegressionModel
# from ...planners.planner import Planner

from dinos.agents.tools.datasets.dataset import Dataset
from dinos.agents.tools.strategies.strategy_set import StrategySet
from dinos.agents.tools.policies.policy import LearningPolicy
from dinos.environments.priors.maps.auto import Mapper


class Learner(Agent):
    """Default Learner, learns by episode but without any choice of task and goal, choose strategy randomly."""

    DATASET_CLASS = Dataset

    def __init__(self, host, dataset=None, performer=None, planner=None, options={}):
        """
        dataset Dataset: dataset of the agent
        strategies Strategy list: list of learning strategies available to the agent
        env Environment: host of the experiment
        """
        dataset = parameter(dataset, self.DATASET_CLASS())
        super().__init__(host, dataset=dataset, performer=performer,
                         planner=planner, options=options)
        self.dataset.attachLearner(self)

        self.learningPolicy = LearningPolicy.EACH_ITERATION
        self.processedEvents = []

        # if self.dataset:
        #     self.addChildModule(self.dataset)

        self.trainStrategies = StrategySet(agent=self)
        # self.reachStrategies.append(reachStrategies if reachStrategies else None)#AutonomousExploration(self))

        self.featureMap = Mapper.get(self.environment.__class__)

    def _serialize(self, serializer):
        dict_ = super()._serialize(serializer)
        dict_.update(serializer.serialize(
            self, ['dataset', 'trainStrategies', 'processedEvents']))
        return dict_
    
    def findStrategy(self, name):
        return next((strategy for strategy in self.trainStrategies if strategy.name == name), None)

    # @classmethod
    # def _deserialize(cls, dict_, environment, dataset=None, options={}, obj=None):
    #     from ...utils.loaders import DataManager

    #     obj = obj if obj else cls(
    #         environment, dataset, options=dict_.get('options', {}))
    #     obj = Agent._deserialize(dict_, environment, options=options, obj=obj)

    #     for strategy in dict_.get('trainStrategies', []):
    #         obj.trainStrategies.add(DataManager.loadType(strategy['path'], strategy['type'])
    #                                 .deserialize(strategy, obj, options=options))
    #     return obj
    
    def addEvent(self, event, config, cost=1., convertToDataset=True):
        if self.dataset and not config.evaluating and event.iteration not in self.processedEvents:
            if convertToDataset:
                event.convertTo(spaceManager=self.dataset, toData=True)
            self.processedEvents.append(event.iteration)
            self._addEvent(event, config, cost)
    
    def _addEvent(self, event, config, cost=1.):
        self.dataset.addEvent(event, cost=cost)
    
    def preProcessEvent(self, event, config, cost=1., convertToDataset=False):
        if self.dataset and event.iteration not in self.processedEvents:
            if convertToDataset:
                event.convertTo(spaceManager=self.dataset, toData=True)
            self._preProcessEvent(event, config, cost)
    
    def _preProcessEvent(self, event, config, cost=1.):
        pass
    
    def addMemory(self, memory, config):
        memory = self._convertMemory(memory)

        for event in memory:
            self.preProcessEvent(event, config)

        if not config.evaluating:
            for event in memory:
                self.addEvent(event, config, convertToDataset=False)
            self.dataset.updated()
    
    def _convertMemory(self, memory):
        memory = list(memory)
        if self.dataset:
            for event in memory:
                event.convertTo(spaceManager=self.dataset, toData=True)
        return memory

    def trainable(self):
        return True

    def train(self, iterations=None, untilIteration=None, episodes=None, untilEpisode=None):
        self.syncCounter()
        self.iterationType[self.iteration] = 'train'
        self.schedule(self._trainSchedule, iterations=iterations,
                      untilIteration=untilIteration, episodes=episodes, untilEpisode=untilEpisode)
    
    def _trainSchedule(self, iterations=None, untilIteration=None, episodes=None, untilEpisode=None):
        """Runs the learner until max number of iterations."""
        goalIteration = untilIteration if untilIteration else (
            self.iteration + iterations if iterations else None)
        goalEpisode = untilEpisode if untilEpisode else (
            self.episode + episodes if episodes else None)
        while ((goalIteration is None or self.iteration < goalIteration) and
               (goalEpisode is None or self.episode < goalEpisode)):
            self.syncCounter()
            self._train()
            self.syncCounter()
        self.iterationType[self.iteration] = 'end'

    def _train(self):
        self.trainEpisode()

    def trainEpisode(self):
        """Run one learning episode."""
        self.environment.setupEpisode()

        self._trainEpisode()

        self.iterationByEpisode.append(
            self.iteration - self.lastIterationEpisode)
        self.lastIterationEpisode = self.iteration
        self.episode += 1
        #self.measureIterationTime()

    def _trainEpisode(self):
        config = self._preEpisode()
        self.configs[self.iteration] = config

        # Performs the episode
        memory = self._performEpisode(config)

        self._postEpisode(memory, config)

    def _performEpisode(self, config):
        # Run an episode of the given strategy
        if config.strategy not in self.trainStrategies:
            raise Exception(f'{config.strategy} is not avaiable within {self}')
        return config.strategy.run(config)

    def _preEpisode(self):
        # Choose learning strategy randomly
        strategy = self.trainStrategies.sample()
        config = MoveConfig(strategy=strategy)

        self.logger.debug(f'Strategy used at iteration {self.iteration}: {config.strategy}', tag='strategy')
        return config

    def _postEpisode(self, memory, config):
        # self.logger.info('Adding episode of length {} to the dataset'
        #                  .format(len(memory)), 'DATA')
        self.addMemory(memory, config)
    
    def _postTest(self, memory, config):
        self.addMemory(memory, config)

    class ResultAnalyser(Agent.ResultAnalyser):
        @property
        def trainingConfigs(self):
            return {i: c for i, c in self.agent.configs.items() if c.training}
        
        def reachedSpaces(self, data, key=None):
            if key is None:
                key = lambda item: item.space
            if isinstance(data, dict):
                data = data.values()
            return list(set([key(item) for item in data]))
        
        def groupBySpace(self, data, key=None):
            if key is None:
                key = lambda item: item.space

            spaces = self.reachedSpaces(data, key=key)
            if isinstance(data, dict):
                return {space: {i: item for i, item in data.items() if key(item) == space} for space in spaces}
            else:
                return {space: [item for item in data if key(item) == space] for space in spaces}

        def _filterNoDict(self, data, iteration=True):
            if not iteration:
                return list(data.values())
            return data
        
        def _filterString(self, data):
            return data and not isinstance(data, str)

        # Analyse
        def randomActions(self, range_=None, iteration=True):
            points = {i: config.result.action for i, config in self.trainingConfigs.items() if config.result.action}
            return self._filterNoDict(points, iteration)

        def goals(self, range_=None, iteration=True):
            points = {i: config.goal for i, config in self.trainingConfigs.items() if config.goal}
            return self._filterNoDict(points, iteration)

        def contextGoals(self, range_=None, iteration=True):
            points = {i: config.goalContext for i, config in self.trainingConfigs.items() if config.goalContext}
            return self._filterNoDict(points, iteration)
        
        def reachedGoals(self, range_=None, iteration=True):
            points = {i: config.result.reachedGoal for i, config in self.trainingConfigs.items() if self._filterString(config.result.reachedGoal)}
            return self._filterNoDict(points, iteration)
        
        def reachedContextGoals(self, range_=None, iteration=True):
            points = {i: config.result.reachedContext for i, config in self.trainingConfigs.items() if self._filterString(config.result.reachedContext)}
            return self._filterNoDict(points, iteration)
        
        def goalErrors(self, range_=None, iteration=True):
            points = {i: (config.goal, config.result.reachedGoal - config.absoluteGoal)
                      for i, config in self.trainingConfigs.items() if config.absoluteGoal and self._filterString(config.result.reachedGoal)}
            return self._filterNoDict(points, iteration)

        # Visual
        def visualizeGeneralData(self, data, title='', color=False, options={}):
            gs = []
            grouped = self.groupBySpace(data)
            for space, data in grouped.items():
                g = Graph(title=f'{title} from {space}', options=options)
                g.scatter(list(data.values()), color=list(data.keys()) if color else None)
                gs.append(g)
            return gs

        def visualizeRandomActions(self, range_=None, color=False, options={}):
            return self.visualizeGeneralData(self.randomActions(range_=range_), title='Random actions', color=color, options=options)

        def visualizeGoals(self, range_=None, color=False, options={}):
            return self.visualizeGeneralData(self.goals(range_=range_), title='Goals', color=color, options=options)
        
        def visualizeContextGoals(self, range_=None, color=False, options={}):
            return self.visualizeGeneralData(self.contextGoals(range_=range_), title='ContextGoals', color=color, options=options)
        
        def visualizeReachedGoals(self, range_=None, color=False, options={}):
            return self.visualizeGeneralData(self.reachedGoals(range_=range_), title='Reached Goals', color=color, options=options)

        def visualizeReachedContextGoals(self, range_=None, color=False, options={}):
            return self.visualizeGeneralData(self.reachedContextGoals(range_=range_), title='Reached ContextGoals', color=color, options=options)

        def visualizeGoalErrors(self, range_=None, color=False, options={}):
            title = 'Goal Errors'
            data = self.goalErrors(range_=range_)

            gs = []
            grouped = self.groupBySpace(data, key=lambda item: item[0].space)
            for space, data in grouped.items():
                values = [(d[0].plain(), d[1].plain()) for d in data.values()]
                g = Graph(title=f'{title} from {space}', options=options)
                g.arrow(values, color=list(data.keys()) if color else None)
                gs.append(g)
            return gs

    # Api
    # def apiget_time(self, =(-1, -1)):
    #     return {'data': iterrange(self.iterationTimes, range_)}



'''
    Author: Alexandre Manoury
    Python Version: 3.6
'''

import numpy as np

from exlab.interface.serializer import Serializable, Serializer
from exlab.utils.text import colorText, Colors
from exlab.utils.ensemble import Ensemble

from dinos.data.data import Observation
from dinos.data.space import SpaceKind

from .entity import Entity


"""

Scene[
    Entity(
        [Observable, ...]
        [Effector, ...]
    ),
...]

"""


class LiveEntity(Entity):
    """Represents a world entity with a set of properties (features and effectors)

    Args:
        kind (string): the type of our entity
        absoluteName (string): Absolute name used to point to this entity. Unique

    Attributes:
        index (int): Absolute index
        indexKind (int): Index for the given entity type
        parent (Entity):
        activated (bool):
        absoluteName
        kind

    """
    PHYSICAL = False
    DISCRETE_STATES = False
    DISCRETE_ACTIONS = False
    CAN_BE_DISCRETIZED = False

    def __init__(self, kind, absoluteName='', disconnected=False, manager=None):
        super().__init__(kind, absoluteName, disconnected, manager=manager)
        self.host = False
        self.hosting = None
        self.scheduledAction = False

        self.physicals = []
        self.actionQueue = []

    # def cid(self):
    #     return self.index

    # def gid(self):
    #     if self.absoluteName:
    #         return Serializer.make_gid(self, self.kind, self.absoluteName)
    #     return Serializer.make_gid(self, self.kind, self.indexKind)

    def _serialize(self, serializer):
        dict_ = serializer.serialize(
            self, ['kind', 'absoluteName', 'index', 'indexKind', 'parent', '_children', '_properties'])
        return dict_

    # @classmethod
    # def _deserialize(cls, dict_, options=None, obj=None):
    #     obj = obj if obj else cls(dict_.get('kind'), dict_.get(
    #         'absoluteName', ''), disconnected=True)
    #     return obj

    def markAsHost(self):
        self.host = True
    
    def hosts(self):
        children = self.cascadingChildren()
        return [host for host in children if host.host]
    
    def findHost(self, name=''):
        if not name:
            return next(iter(self.hosts()), None)
        return next((host for host in self.hosts() if host.absoluteName == name), None)

    def _update(self, dt):
        pass

    def draw(self, *args, **kwargs):
        pass

    def update(self, dt=0.):
        self._update(dt)
        for action in list(self.actionQueue):
            if action.update(dt):
                self.actionQueue.remove(action)
        for child in self._children:
            child.update(dt)

    def performAction(self, function, args=(), duration=0., step=None):
        self.actionQueue.append(ActionExecution(
            function, args, duration, step))


class ActionExecution(object):
    TOLERANCE_STEP = 1e-5

    def __init__(self, function, args=(), duration=0., step=None):
        self.function = function
        self.args = args
        self.duration = duration
        self.step = step
        self.elapsed = 0.
        self.lastExecution = -1000.

    def update(self, dt):
        if not self.step or self.elapsed - self.lastExecution >= self.step - self.TOLERANCE_STEP:
            self._perform()
            self.lastExecution = self.elapsed

        self.elapsed += dt
        if self.elapsed >= self.duration:
            return True
        return False

    def _perform(self):
        self.function(*self.args)



from .internal_engine import InternalEngine

from panda3d.core import Vec3
from panda3d.bullet import BulletWorld
from panda3d.bullet import BulletPlaneShape
from panda3d.bullet import BulletRigidBodyNode
from panda3d.bullet import BulletBoxShape
from pandac.PandaModules import ClockObject
from pandac.PandaModules import loadPrcFileData


class PandaEnvironment(InternalEngine):
    ENGINE = 'Panda3D'

    def __init__(self, scene=None, options={}):
        super().__init__(scene, options)

        self.jump_frame = self.speedup_frames
        self.jump_physics_frame = 0

    def initPhysics(self):
        # Physics stuff
        # World
        self.space = BulletWorld()
        self.space.setGravity(Vec3(0, 0, -9.81))

        # Plane
        shape = BulletPlaneShape(Vec3(0, 0, 1), 1)
        node = BulletRigidBodyNode('Ground')
        node.addShape(shape)
        np = render.attachNewNode(node)
        np.setPos(0, 0, -2)
        self.space.attachRigidBody(node)

    def setGui(self, gui):
        if self.gui == gui:
            return
        self.gui = gui
        if self.gui:
            loadPrcFileData("", "window-title Your Title")
            loadPrcFileData("", "win-size 800 600")
            loadPrcFileData("", "sync-video 0")
        else:
            loadPrcFileData("", "window-type none")

        from direct.showbase import ShowBase
        base = ShowBase.ShowBase()

        if self.gui:
            base.cam.setPos(0, -10, 0)
            base.cam.lookAt(0, 0, 0)
            base.setFrameRateMeter(True)

    def reset(self):
        for obj in self.agents + self.objs:
            pass  # TODO

    def _updatePhysics(self, dt):
        self.jump_physics_frame += 1
        if self.jump_physics_frame >= 5:
            self.space.doPhysics(0.001)
            self.jump_physics_frame = 0

    def flipScreen(self):
        self.jump_frame += 1.
        if self.jump_frame >= self.speedup_frames:
            taskMgr.step()
            self.jump_frame = 0.

    def customStep(self, duration):
        '''if self.gui and duration is None:
            for event in pygame.event.get():
                if event.type == QUIT:
                    self.running = False
                elif event.type == KEYDOWN and event.key == K_ESCAPE:
                    self.running = False
                elif event.type == KEYDOWN and event.key == K_j:
                    # self.agents[0].execute_action((1.0, 0.0))
                    # change mode
                    if self.current_space == self.env.dataset.outcomeSpaces[-1]:
                        self.current_space = self.env.dataset.outcomeSpaces[0]
                    else:
                        self.current_space = self.env.dataset.outcomeSpaces[self.env.dataset.outcomeSpaces.index(self.current_space) + 1]
                elif event.type == pygame.MOUSEBUTTONUP:
                    # convert to local agent coordinates
                    pos = self.propertyItem(self.current_space.property).convert_to_feature(pymunk.pygame_util.from_pygame(event.pos, self.screen))
                    self.experiment.exploit(pos, self.env.dataset.outcomeSpaces.index(self.current_space))'''

        # Draw stuff
        '''if self.gui:
            self.screen.fill(THECOLORS["white"])  # Clear screen
            self.space.debug_draw(self.draw_options)
            if duration:
                label_acting = self.font.render("Acting", 1, (255, 0, 0))
                self.screen.blit(label_acting, (24, 24))
            else:
                label_mode = self.font.render("Current outcome space: " + self.current_space.name + " [J] to change", 1,
                                              (0, 0, 0))
                self.screen.blit(label_mode, (24, 24))'''



import bz2
import sys
import copy
import math
import pickle
import random
import numpy as np

from enum import Enum
from scipy.spatial.distance import euclidean
from sklearn.neighbors import NearestNeighbors

from exlab.interface.serializer import Serializable
from exlab.interface.graph import Graph

from dinos.data.data import SingleData, Data, Goal, Action
from .space import Space, SpaceKind

# from ..utils.io import getVisual, plotData, visualize
from . import operations


class DataSpace(Space):
    """
    Abstract class representing a physical space
    """

    def __init__(self, spaceManager, dim, options={}, native=None, kind=SpaceKind.BASIC, spaces=None):
        super().__init__(spaceManager, dim, options=options,
                         native=native, kind=kind, spaces=spaces)

        self._number = 0
        self.clearSpaceWeight()

        self.costs = np.zeros([Space.RESERVE_STEP])
        self.data = np.zeros([Space.RESERVE_STEP, self.dim])
        # execution id (order)
        self.ids = np.zeros([Space.RESERVE_STEP], dtype=np.int32)
        # reverse execution order
        self.lids = np.full([Space.RESERVE_STEP], -1, dtype=np.int32)
        self.actions = []

        self.contiguous = True  # ids == lids
        self.incoherentRowsAlignement = False

        self._nnWeightsCache = {}
    
    def _serialize(self, serializer):
        dict_ = super()._serialize(serializer)
        dict_.update(serializer.serialize(
            self, ['_number', 'contiguous', 'incoherentRowsAlignement']))
        for attr in ['costs', 'data', 'ids']:
            dict_[attr] = bz2.compress(pickle.dumps(getattr(self, attr)[:self._number]))
        a = np.argwhere([self.lids != -1])
        if len(a) == 0:
            n = 0
        else:
            n = np.max(a[:, 1])
        dict_['lids'] = bz2.compress(pickle.dumps(self.lids[:n+1]))
        return dict_
    
    # @classmethod
    # def _deserialize(cls, dict_, serializer, obj=None):
    #     if obj is None:
    #         obj = serializer.get('environment').propertySpace(serializer.get(dict_.get('boundProperty')['__id__']),
    #                                                           SpaceKind(dict_.get('kind')),
    #                                                           serializer.get('dataset'))
    #     return super()._deserialize(dict_, serializer, obj)

    def _postDeserialize(self, dict_, serializer):
        super()._postDeserialize(dict_, serializer)
        self._number = dict_.get('number')
        self.contiguous = dict_.get('contiguous')
        self.incoherentRowsAlignement = dict_.get('incoherentRowsAlignement')

        for attr in ['costs', 'data', 'ids', 'lids']:
            if attr in dict_:
                d = pickle.loads(bz2.decompress(dict_.get(attr)))
                getattr(self, attr)[:len(d)] = d
        
        self._validate()
        self._updateSpaceWeights()

    def icon(self):
        return '@☰'

    def canStoreData(self):
        return True

    @property
    def spaceWeights(self):
        return self._spaceWeights

    def _updateSpaceWeights(self):
        if not self._spaceWeights:
            self._nnWeights = np.ones(self.dim, dtype=np.float) / self.dim
        else:
            weights = np.array(self._spaceWeights)
            weights[:, 1] *= sum([space.maxDistance for space in weights[:, 0]]) / sum(weights[:, 1])
            weights[:, 1] = [weight / space.maxDistance / space.dim for space, weight in weights]
            self._nnWeights = np.zeros(len(self.spaces), dtype=np.float)

            def getWeight(space):
                if space not in weights[:, 0]:
                    return 0.
                return [weight for sp, weight in weights if sp == space][0]

            weightPerSpace = [
                [getWeight(space)] * space.dim for space in self.spaces]
            self._nnWeights = np.array(
                [weight for wlist in weightPerSpace for weight in wlist], dtype=np.float)

    def clearSpaceWeight(self):
        self._spaceWeights = []
        self._updateSpaceWeights()

    def spaceWeight(self, weight, space=None):
        if weight is None:
            self.clearSpaceWeight()
        else:
            if isinstance(weight, list):
                self._spaceWeights = weight
            elif space is None:
                self._spaceWeights = [(space, weight) for space in self.spaces]
            else:
                weight /= len(space.spaces)
                for subspace in space:
                    self._spaceWeights.append((subspace, weight))
            self._updateSpaceWeights()

    # Data
    @property
    def number(self):
        self._validate()
        return self._number

    def getLid(self, ids):
        self._validate()
        if self.contiguous:
            return ids
        a = self.lids[ids]
        return a[a >= 0]

    def getIds(self, restrictionIds=None):
        self._validate()
        return self.ids[:self._number]

    def __getitem__(self, ids):
        self._validate()
        return self.data[self.getLid(ids)]

    def getPoint(self, ids, toSpace=None):
        self._validate()
        a = self.lids[ids]
        data = self.data[a[a >= 0]].tolist()
        return [self.point(d, toSpace=toSpace) for d in data]

    def getNpPlainPoint(self, ids):
        self._validate()
        a = self.lids[ids]
        return self.data[a[a >= 0]]

    def getPlainPoint(self, ids):
        return self.getNpPlainPoint(ids).tolist()

    def getActionIndex(self, ids):
        self._validate()
        return list(set.intersection(set(ids), set(self.actions)))

    def getNpPlainAction(self, ids):
        ids = self.getActionIndex(ids)
        if not ids:
            return np.array([])
        return self.data[ids]

    def getPlainAction(self, ids):
        list_ = self.getNpPlainAction(ids)
        return [a.tolist() for a in list_]

    # Operations
    def findWeights(self, columns=None, adjustSpaceWeight=None, adjustWeightFactor=1., weights=None):
        # weights = np.array(self._nnWeights) if weights is None else self._nnWeights * weights
        if adjustSpaceWeight and columns is not None and np.any(columns):
            key = (adjustSpaceWeight, adjustWeightFactor, tuple(columns))
            if key in self._nnWeightsCache:
                newWeights = self._nnWeightsCache[key] * self._nnWeights
            else:
                newWeights = np.ones(self.dim)
                adjustColumns = self.columnsFor(adjustSpaceWeight)
                if np.any(adjustColumns):
                    newWeights[adjustColumns] *= adjustWeightFactor * len(adjustColumns) / np.sum(columns[adjustColumns])
                self._nnWeightsCache[key] = newWeights
                newWeights = newWeights * self._nnWeights
        else:
            newWeights = self._nnWeights
        return newWeights if weights is None else newWeights * weights
    
    # def findWeights(self, columns=None, adjustSpaceWeight=None, adjustWeightFactor=1., weights=None):
    #     weights = np.array(self._nnWeights) if weights is None else self._nnWeights * weights
    #     if adjustSpaceWeight and columns is not None and np.any(columns):
    #         key = (adjustSpaceWeight, adjustWeightFactor, columns)
    #         if key in self._nnWeightsCache:
    #             return self._nnWeightsCache[key]
    #         adjustColumns = self.columnsFor(adjustSpaceWeight)
    #         if np.any(adjustColumns):
    #             weights[adjustColumns] *= adjustWeightFactor * len(adjustColumns) / np.sum(columns[adjustColumns])
    #     return weights

    def __computeDistances(self, x, weights=None, columns=None, restrictionLids=None, adjustSpaceWeight=None, adjustWeightFactor=1.):
        """Compute array of normalized distances between data and the point given."""
        data = self.data[:self._number]
        if restrictionLids is not None:
            data = data[restrictionLids]
        weights = self.findWeights(columns, adjustSpaceWeight, adjustWeightFactor, weights)
        # print(data)
        # print(weights)
        try:
            return operations.euclidean_distances(data, x, weights, self.maxDistance, columns)
        except Exception as e:
            raise Exception(f'Failure __computeDistances: {data.dtype} {x.dtype} {weights.dtype}\n{self} {x} {weights} {self.maxDistance} {columns}\n{e}')

    def __computePerformances(self, x, weights=None, columns=None, restrictionLids=None, adjustSpaceWeight=None, adjustWeightFactor=1.):
        """Compute performances for reaching the given point."""
        costs = self.costs[:self._number]
        if restrictionLids is not None:
            costs = costs[restrictionLids]
        return self.__computeDistances(x, weights, columns, restrictionLids, adjustSpaceWeight, adjustWeightFactor) * costs

    def __nearestNeighbors(self, x, n=1, ignore=0, weights=None, columns=None, restrictionIds=None, otherSpace=None,
                           useDistances=True, adjustSpaceWeight=None, adjustWeightFactor=1.):
        self._validate()
        if self._number == 0:
            return np.array([], dtype=np.int32), np.array([])
        if otherSpace:
            otherSpace._validate()
            if otherSpace._number == 0:
                return np.array([], dtype=np.int32), np.array([])

        restrictionLids = operations.findRestrictionsLids(restrictionIds, self, otherSpace)
        if useDistances:
            data = self.__computeDistances(x, weights, columns, restrictionLids, adjustSpaceWeight, adjustWeightFactor)
        else:
            data = self.__computePerformances(x, weights, columns, restrictionLids, adjustSpaceWeight, adjustWeightFactor)
        return operations.nearestNeighbors(self.ids, data, n, ignore, restrictionLids, otherSpace)

    def nearest(self, x, n=1, ignore=0, restrictionIds=None, otherSpace=None, weights=None, columns=None, adjustSpaceWeight=None, adjustWeightFactor=1.):
        """Computes Nearest Neighbours based on performances."""
        return self.__nearestNeighbors(x.npPlain(), n=n, ignore=ignore, weights=weights, columns=columns, restrictionIds=restrictionIds,
                                       otherSpace=otherSpace, adjustSpaceWeight=adjustSpaceWeight, adjustWeightFactor=adjustWeightFactor, useDistances=False)

    def nearestDistance(self, x, n=1, ignore=0, restrictionIds=None, otherSpace=None, weights=None, columns=None, adjustSpaceWeight=None, adjustWeightFactor=1.):
        """Compute Nearest Neighbours based on distance."""
        return self.__nearestNeighbors(x.npPlain(), n=n, ignore=ignore, weights=weights, columns=columns, restrictionIds=restrictionIds,
                                       otherSpace=otherSpace, adjustSpaceWeight=adjustSpaceWeight, adjustWeightFactor=adjustWeightFactor)

    def nearestPlain(self, x, n=1, ignore=0, restrictionIds=None, otherSpace=None, weights=None, columns=None, adjustSpaceWeight=None, adjustWeightFactor=1.):
        """Computes Nearest Neighbours based on performances."""
        return self.__nearestNeighbors(x, n=n, ignore=ignore, weights=weights, columns=columns, restrictionIds=restrictionIds,
                                       otherSpace=otherSpace, adjustSpaceWeight=adjustSpaceWeight, adjustWeightFactor=adjustWeightFactor, useDistances=False)

    def nearestDistancePlain(self, x, n=1, ignore=0, restrictionIds=None, otherSpace=None, weights=None, columns=None, adjustSpaceWeight=None, adjustWeightFactor=1.):
        """Compute Nearest Neighbours based on distance."""
        return self.__nearestNeighbors(x, n=n, ignore=ignore, weights=weights, columns=columns, restrictionIds=restrictionIds,
                                       otherSpace=otherSpace, adjustSpaceWeight=adjustSpaceWeight, adjustWeightFactor=adjustWeightFactor)

    @staticmethod
    def nearestFromData(points, x, n=1, ignore=0):
        return operations.nearestNeighborsFromData(points, x, n, ignore)
    
    @staticmethod
    def nearestFromDataContiguous(points, x, n=1, ignore=0, columns=None):
        return operations.nearestNeighborsFromDataContiguous(points, x, n, ignore, columns)

    def nearestDistanceArray(self, x, n=1, ignore=0, restrictionIds=None, otherSpace=None, weights=None, columns=None):
        self._validate()
        if self._number == 0:
            return np.array([], dtype=np.int32), np.array([])
        if otherSpace:
            otherSpace._validate()
            if otherSpace._number == 0:
                return np.array([], dtype=np.int32), np.array([])

        weights = self.maxDistancePerColumn
        data = self.data[:self._number]
        ids = self.ids[:self._number]

        restrictionLids = operations.findRestrictionsLids(restrictionIds, self, otherSpace)
        if restrictionLids is not None:
            data = data[restrictionLids]
            ids = ids[restrictionLids]

        if len(ids) == 0:
            return np.array([], dtype=np.int32), np.array([])

        x = np.array(x)
        weights = self._nnWeights if weights is None else self._nnWeights * weights
        if columns is not None:
            data = data[:, columns]
            x = x[:, columns]
            weights = weights[columns]
        data = data * weights
        x = x * weights

        nbrs = NearestNeighbors(n_neighbors=min(
            n + ignore, data.shape[0]), algorithm='ball_tree').fit(data)
        distances, indices = nbrs.kneighbors(x)
        return ids[indices[:, ignore:]], distances[:, ignore:]

    def variance(self, ids):
        points = self.getPlainPoint(ids)
        center = np.mean(points, axis=0)
        return np.mean(np.sum((points - center) ** 2, axis=1) ** .5)

    def denseEnough(self, ids, threshold=0.05):
        variance = self.variance(ids)
        return variance < self.maxDistance * threshold

    def getData(self, restrictionIds=None):
        self._validate()
        if restrictionIds is not None:
            try:
                return self.data[self.getLid(restrictionIds)]
            except Exception as e:
                print(self.ids)
                print(restrictionIds)
                raise e
        return self.data[:self._number]

    def getDataSelection(self, n=0, method='first', restrictionIds=None):
        data = self.getData(restrictionIds=restrictionIds)
        if not n:
            n = len(data)
        if method == 'random':
            data[np.random.choice(data.shape[0], n)]
        elif method == 'first':
            data = data[:n]
        return data
    
    # def findDuplicates(self, starting=-1):
    #     self._validate()
    #     if self.lastDuplicateSearch >= self._number:
    #         return

    #     self.noDuplicates = 

    #     self.lastDuplicateSearch = self._number

    def addPoint(self, point, idx, cost=None, action=False):
        """Add a point in the space and if valid return id."""
        if point.space.nativeRoot != self.nativeRoot:
            return -1

        x = point.plain()

        if self._number > 0 and self.ids[self._number - 1] == idx:
            raise Exception(f'Trying to add data twice to {self} at index {idx}:\n1: {self.data[self._number - 1]}\n2: {x}')

        # Extend arrays if needed
        if self._number >= self.data.shape[0]:
            self.data = np.append(self.data, np.zeros(
                [Space.RESERVE_STEP, self.dim]), axis=0)
            self.costs = np.append(self.costs, np.zeros(
                [Space.RESERVE_STEP]), axis=0)
            self.ids = np.append(self.ids, np.zeros(
                [Space.RESERVE_STEP], dtype=np.int16), axis=0)
        if idx >= self.lids.shape[0]:
            self.lids = np.append(self.lids, np.full(
                [Space.RESERVE_STEP], -1, dtype=np.int16), axis=0)

        # Store data
        self.data[self._number] = x
        self.costs[self._number] = cost if cost else 1.
        self.ids[self._number] = idx
        self.lids[idx] = self._number
        if idx != self._number:
            self.contiguous = False
        self._number += 1
        if action:
            self.actions.append(idx)

        self.invalidate()
        return self._number - 1

    def _postValidate(self):
        if self._number > 0:
            self._bounds = list(zip(np.min(self.data[:self._number], axis=0).tolist(),
                                    np.max(self.data[:self._number], axis=0).tolist()))
            self.maxDistance = math.sqrt(
                sum([(bound[1] - bound[0]) ** 2 for bound in self._bounds]))
        else:
            self.maxDistance = 1.
        if self.maxDistance == 0:
            self.maxDistance = 1.
        if self.aggregation:
            for space in self.spaces:
                self.maxDistancePerColumn[self.columnsFor(space)] = space.maxDistance
        else:
            self.maxDistancePerColumn = np.full(self.dim, self.maxDistance)
        self.maxNNDistance = self.maxDistance
        Space._postValidate(self)

    # @classmethod
    # def getCost(cls, n):
    #     """Return cost of an action space based on its _number of primitives."""
    #     return cls.gamma ** n

    # Visual
    def visualizeData(self, options={}):
        g = Graph(title=f'Points from {self}', options=options)
        g.scatter(self.getData())
        return g

    # def getPointsVisualizer(self, prefix=""):
    #     """Return a dictionary used to visualize outcomes reached for the specified outcome space."""
    #     return getVisual(
    #         [lambda fig, ax, options: plotData(
    #             self.getData(), fig, ax, options)],
    #         minimum=[b[0] for b in self._bounds],
    #         maximum=[b[1] for b in self._bounds],
    #         title=prefix + "Points in " + str(self)
    #     )

    # def plot(self):
    #     visualize(self.getPointsVisualizer())

    # # Api
    # def apiGetPoints(self, ids):
    #     self._validate()
    #     # if range_[1] == -1:
    #     #     ids = np.nonzero(self.ids[:self._number] > range_[0])
    #     # else:
    #     #     ids = np.nonzero(np.logical_and(self.ids[:self._number] > range_[0], self.ids[:self._number] <= range_[1]))
    #     # print(self.ids[ids].size)
    #     # print(self.data[ids].size)
    #     # data = np.concatenate((self.ids[ids], self.data[ids]), axis=1)
    #     return list(zip(self.data[ids].tolist(), self.ids[ids].tolist()))
